@@ -1,13 +1,14 @@
 from . import pql_objects as objects
 from . import sql
 from .pql_types import T, dp_type, ITEM_NAME
+from .types_impl import kernel_type 
 from .exceptions import Signal
 
 @dp_type
 def _cast(state, inst_type, target_type, inst):
     if inst_type == target_type or target_type is T.any:
         return inst
-    raise Signal.make(T.TypeError, state, None, f"Cast not implemented for {inst_type}->{target_type}")
+    raise Signal.make(T.TypeError, None, f"Cast not implemented for {inst_type}->{target_type}")
 
 @dp_type
 def _cast(state, inst_type: T.list, target_type: T.list, inst):
@@ -32,14 +33,14 @@ def _cast(state, inst_type: T.aggregate, target_type: T.list, inst):
 def _cast(state, inst_type: T.table, target_type: T.list, inst):
     t = inst.type
     if len(t.elems) != 1:
-        raise Signal.make(T.TypeError, state, None, f"Cannot cast {inst_type} to {target_type}. Too many columns")
+        raise Signal.make(T.TypeError, None, f"Cannot cast {inst_type} to {target_type}. Too many columns")
     if not (inst_type.elem <= target_type.elem):
-        raise Signal.make(T.TypeError, state, None, f"Cannot cast {inst_type} to {target_type}. Elements not matching")
+        raise Signal.make(T.TypeError, None, f"Cannot cast {inst_type} to {target_type}. Elements not matching")
 
     (elem_name, elem_type) ,= inst_type.elems.items()
     code = sql.Select(T.list[elem_type], inst.code, [sql.ColumnAlias(sql.Name(elem_type, elem_name), ITEM_NAME)])
 
-    return objects.ListInstance.make(code, T.list[elem_type], [inst])
+    return objects.TableInstance.make(code, T.list[elem_type], [inst])
 
 @dp_type
 def _cast(state, inst_type: T.t_id, target_type: T.int, inst):
@@ -51,11 +52,7 @@ def _cast(state, inst_type: T.int, target_type: T.t_id, inst):
 
 @dp_type
 def _cast(state, inst_type: T.union[T.float, T.bool], target_type: T.int, inst):
-    if state.db.target is sql.mysql:
-        t = "signed integer"
-    else:
-        t = "int"
-    code = sql.Cast(T.int, t, inst.code)
+    code = sql.Cast(T.int, inst.code)
     return objects.Instance.make(code, T.int, [inst])
 
 @dp_type
@@ -80,7 +77,7 @@ def _cast(state, inst_type: T.string, target_type: T.string, inst):     # Disamb
 
 @dp_type
 def _cast(state, inst_type: T.union[T.int, T.bool], target_type: T.float, inst):
-    code = sql.Cast(T.float, "float", inst.code)
+    code = sql.Cast(T.float, inst.code)
     return objects.Instance.make(code, T.float, [inst])
 
 # @dp_type
@@ -91,7 +88,7 @@ def _cast(state, inst_type: T.union[T.int, T.bool], target_type: T.float, inst):
 
 @dp_type
 def _cast(state, inst_type: T.primitive, target_type: T.string, inst):
-    code = sql.Cast(T.string, "varchar", inst.code)
+    code = sql.Cast(T.string, inst.code)
     return objects.Instance.make(code, T.string, [inst])
 
 @dp_type
@@ -106,8 +103,5 @@ def _cast(state, inst_type: T.t_relation, target_type: T.int, inst):
     raise Signal.make(T.TypeError, state, None, f"Cast not implemented for {inst_type}->{target_type}")
 
 def cast(state, obj, t):
-    was_vec, [obj, t] = objects.unvectorize_args([obj, t])
-    res = _cast(state, obj.type, t, obj)
-    if was_vec:
-        res = objects.vectorized(res)
-    return res
\ No newline at end of file
+    res = _cast(state, kernel_type(obj.type), t, obj)
+    return objects.inherit_vectorized(res, [obj])
\ No newline at end of file
