@@ -1,7 +1,5 @@
 import operator
 
-from runtype import DispatchError
-
 from .utils import safezip, listgen, find_duplicate, SafeDict, re_split
 from .exceptions import Signal
 from . import exceptions as exc
@@ -11,10 +9,10 @@ from . import pql_objects as objects
 from . import pql_ast as ast
 from . import sql
 from .interp_common import dy, State, assert_type, new_value_instance, evaluate, simplify, call_pql_func, cast_to_python
-from .pql_types import T, Type, union_types, Id, ITEM_NAME, dp_inst
-from .types_impl import flatten_type, pql_repr
+from .pql_types import T, Type, Id, ITEM_NAME, dp_inst
+from .types_impl import flatten_type, pql_repr, kernel_type
 from .casts import cast
-from .pql_objects import AbsInstance, vectorized, unvectorized, make_instance
+from .pql_objects import AbsInstance, vectorized, make_instance, inherit_vectorized_type, unvectorized
 
 class AutocompleteSuggestions(Exception):
     pass
@@ -30,7 +28,7 @@ def cast_to_instance(state, x):
         inst = compile_to_inst(state, x)
         # inst = evaluate(state, x)
     except exc.ReturnSignal:
-        raise Signal.make(T.CompileError, state, None, f"Bad compilation of {x}")
+        raise Signal.make(T.CompileError, None, f"Bad compilation of {x}")
 
     if isinstance(inst, ast.ParameterizedSqlCode):
         raise exc.InsufficientAccessLevel(inst)
@@ -38,7 +36,7 @@ def cast_to_instance(state, x):
     if not isinstance(inst, AbsInstance):
         # TODO compile error? cast error?
         # TODO need to be able to catch this above, and provide better errors
-        raise Signal.make(T.TypeError, state, None, f"Could not compile {pql_repr(state, inst.type, inst)}")
+        raise Signal.make(T.TypeError, None, f"Could not compile {pql_repr(inst.type, inst)}")
 
     return inst
 
@@ -63,7 +61,7 @@ def _process_fields(state: State, fields):
 
         if v.type <= T.aggregate:
             v = v.primary_key()
-            t = T.list[v.type]
+            t = T.json_array[v.type]
             v = make_instance(sql.MakeArray(t, v.code), t, [v])
 
         suggested_name = str(f.name) if f.name else guess_field_name(f.value)
@@ -90,11 +88,9 @@ def _expand_ellipsis(state, obj, fields):
 
         if isinstance(f.value, ast.Ellipsis):
             if f.name:
-                raise Signal.make(T.SyntaxError, state, f, "Cannot use a name for ellipsis (inlining operation doesn't accept a name)")
+                raise Signal.make(T.SyntaxError, f, "Cannot use a name for ellipsis (inlining operation doesn't accept a name)")
             else:
-                t = obj.type
-                if t <= T.vectorized:
-                    t = t.elem
+                t = kernel_type(obj.type)
                 assert t <= T.table or t <= T.struct # some_table{ ... } or some_table{ some_struct_item {...} }
 
                 for n in f.value.exclude:
@@ -104,7 +100,7 @@ def _expand_ellipsis(state, obj, fields):
                                                       and k not in f.value.exclude})
 
                     if n in direct_names:
-                        raise Signal.make(T.NameError, state, n, f"Field to exclude '{n}' is explicitely included in projection")
+                        raise Signal.make(T.NameError, n, f"Field to exclude '{n}' is explicitely included in projection")
 
                 if f.value.from_struct:
                     # Inline struct
@@ -119,7 +115,7 @@ def _expand_ellipsis(state, obj, fields):
                     remaining_items = list(_exclude_items(items, set(f.value.exclude), direct_names))
                 except ValueError as e:
                     fte = set(e.args[0])
-                    raise Signal.make(T.NameError, state, obj, f"Fields to exclude '{fte}' not found")
+                    raise Signal.make(T.NameError, obj, f"Fields to exclude '{fte}' not found")
 
                 exclude = direct_names | set(f.value.exclude)
                 for name, value in remaining_items:
@@ -144,14 +140,18 @@ def compile_to_inst(state: State, cb: ast.CodeBlock):
         return compile_to_inst(state, cb.statements[0])
 
     # TODO some statements can be evaluated at compile time
-    raise Signal.make(T.CompileError, state, cb, "Cannot compile this code block")
+    raise Signal.make(T.CompileError, cb, "Cannot compile this code block")
 @dy
 def compile_to_inst(state: State, i: ast.If):
     cond = cast(state, cast_to_instance(state, i.cond), T.bool)
     then = cast_to_instance(state, i.then)
     else_ = cast_to_instance(state, i.else_)
     code = sql.Case(cond.code, then.code, else_.code)
-    return make_instance(code, T.bool, [cond, then, else_])
+    # TODO simplify this with a better type system
+    res_type = kernel_type(then.type) | kernel_type(else_.type)
+    res_type = inherit_vectorized_type(res_type, [cond, then, else_])
+    return make_instance(code, res_type, [cond, then, else_])
+
 
 
 @dy
@@ -162,15 +162,15 @@ def compile_to_inst(state: State, proj: ast.Projection):
         return table   # Empty list projection is always an empty list.
 
     t = T.union[T.collection, T.struct]
-    if not (table.type <= T.union[t, T.vectorized[t]]):
-        raise Signal.make(T.TypeError, state, proj, f"Cannot project objects of type {table.type}")
+    if not (kernel_type(table.type) <= t):
+        raise Signal.make(T.TypeError, proj, f"Cannot project objects of type {table.type}")
 
     fields = _expand_ellipsis(state, table, proj.fields)
 
     # Test duplicates in field names. If an automatic name is used, collision should be impossible
     dup = find_duplicate([f for f in list(proj.fields) + list(proj.agg_fields) if f.name], key=lambda f: f.name)
     if dup:
-        raise Signal.make(T.TypeError, state, dup, f"Field '{dup.name}' was already used in this projection")
+        raise Signal.make(T.TypeError, dup, f"Field '{dup.name}' was already used in this projection")
 
     attrs = table.all_attrs()
 
@@ -178,9 +178,8 @@ def compile_to_inst(state: State, proj: ast.Projection):
         fields = _process_fields(state, fields)
 
     for name, f in fields:
-        t = T.union[T.primitive, T.struct, T.nulltype, T.unknown]
-        if not (f.type <= T.union[t, T.vectorized[t]]):
-            raise exc.Signal.make(T.TypeError, state, proj, f"Cannot project values of type: {f.type}")
+        if not kernel_type(f.type) <= T.union[T.primitive, T.struct, T.json, T.nulltype, T.unknown]:
+            raise exc.Signal.make(T.TypeError, proj, f"Cannot project values of type: {f.type}")
 
     if isinstance(table, objects.StructInstance):
         d = {n[1]:c for n, c in fields}     # Remove used_defined bool
@@ -203,8 +202,7 @@ def compile_to_inst(state: State, proj: ast.Projection):
     elems = {}
     for (user_defined, name), type_ in field_types:
         # Unvectorize for placing in the table type
-        if type_ <= T.vectorized:
-            type_ = type_.elem
+        type_ = kernel_type(type_)
 
         # Find name without collision
         if not user_defined:
@@ -232,7 +230,7 @@ def compile_to_inst(state: State, proj: ast.Projection):
     ]
 
     if not sql_fields:
-        raise Signal.make(T.TypeError, state, proj, "No column provided for projection (empty projection)")
+        raise Signal.make(T.TypeError, proj, "No column provided for projection (empty projection)")
 
     # Make Instance
     new_table = objects.TableInstance.make(sql.null, new_table_type, [table] + [inst for _, inst in all_fields])
@@ -265,7 +263,7 @@ def compile_to_inst(state: State, order: ast.Order):
     for f in fields:
         if not f.type <= T.primitive:
             # TODO Support 'ordering' trait?
-            raise Signal.make(T.TypeError, state, order, f"Arguments to 'order' must be primitive")
+            raise Signal.make(T.TypeError, order, f"Arguments to 'order' must be primitive")
 
 
     code = sql.table_order(table, [c.code for c in fields])
@@ -284,17 +282,12 @@ def compile_to_inst(state: State, lst: list):
     return [evaluate(state, e) for e in lst]
 
 
-def base_type(t):
-    if t <= T.vectorized:
-        return t.elem
-    return t
-
 @dy
 def compile_to_inst(state: State, o: ast.Or):
     args = cast_to_instance(state, o.args)
     a, b = args
-    if base_type(a.type) != base_type(b.type):
-        raise Signal.make(T.TypeError, state, o, f"'or' operator requires both arguments to be of the same type, got {a.type} and {b.type}")
+    if kernel_type(a.type) != kernel_type(b.type):
+        raise Signal.make(T.TypeError, o, f"'or' operator requires both arguments to be of the same type, got {a.type} and {b.type}")
     cond = cast(state, a, T.bool)
     code = sql.Case(cond.code, a.code, b.code)
     return objects.make_instance(code, a.type, args)
@@ -303,8 +296,8 @@ def compile_to_inst(state: State, o: ast.Or):
 def compile_to_inst(state: State, o: ast.And):
     args = cast_to_instance(state, o.args)
     a, b = args
-    if base_type(a.type) != base_type(b.type):
-        raise Signal.make(T.TypeError, state, o, f"'and' operator requires both arguments to be of the same type, got {a.type} and {b.type}")
+    if kernel_type(a.type) != kernel_type(b.type):
+        raise Signal.make(T.TypeError, o, f"'and' operator requires both arguments to be of the same type, got {a.type} and {b.type}")
     cond = cast(state, a, T.bool)
     code = sql.Case(cond.code, b.code, a.code)
     return objects.make_instance(code, a.type, args)
@@ -319,6 +312,11 @@ def compile_to_inst(state: State, o: ast.Not):
 
 
 ## Contains
+def contains(state, op, a, b):
+    res = _contains(state, op, unvectorized(a), unvectorized(b))
+    return objects.inherit_vectorized(res, [a, b])
+
+
 @dp_inst
 def _contains(state, op, a: T.string, b: T.string):
     f = {
@@ -332,7 +330,7 @@ def _contains(state, op, a: T.primitive, b: T.collection):
     b_list = cast(state, b, T.list)
     if not (a.type <= b_list.type.elem):
         a = cast(state, a, b_list.type.elem)
-        # raise Signal.make(T.TypeError, state, op, f"Error in contains: Mismatch between {a.type} and {b.type}")
+        # raise Signal.make(T.TypeError, op, f"Error in contains: Mismatch between {a.type} and {b.type}")
 
     if op == '!in':
         op = 'not in'
@@ -341,29 +339,17 @@ def _contains(state, op, a: T.primitive, b: T.collection):
 
 @dp_inst
 def _contains(state, op, a: T.any, b: T.any):
-    raise Signal.make(T.TypeError, state, op, f"Contains not implemented for {a.type} and {b.type}")
+    raise Signal.make(T.TypeError, op, f"Contains not implemented for {a.type} and {b.type}")
 
-@dp_inst
-def _contains(state, op, a: T.vectorized, b: T.any):
-    return vectorized(_contains(state, op, unvectorized(a), b))
 
 ## Compare
+def compare(state, op, a, b):
+    res = _compare(state, op, unvectorized(a), unvectorized(b))
+    return objects.inherit_vectorized(res, [a, b])
+
 @dp_inst
 def _compare(state, op, a: T.any, b: T.any):
-    raise Signal.make(T.TypeError, state, op, f"Compare not implemented for {a.type} and {b.type}")
-
-@dp_inst
-def _compare(state, op, a: T.vectorized, b: T.vectorized):
-    return vectorized(_compare(state, op, unvectorized(a), unvectorized(b)))
-
-@dp_inst
-def _compare(state, op, a: T.vectorized, b: T.any):
-    return vectorized(_compare(state, op, unvectorized(a), b))
-
-@dp_inst
-def _compare(state, op, a: T.any, b: T.vectorized):
-    return vectorized(_compare(state, op, a, unvectorized(b)))
-
+    raise Signal.make(T.TypeError, op, f"Compare not implemented for {a.type} and {b.type}")
 
 
 @dp_inst
@@ -443,15 +429,15 @@ def _compare(state, op, a: T.type, b: T.type):
     if op == '<=':
         return call_pql_func(state, "issubclass", [a, b])
     if op != '=':
-        raise exc.Signal.make(T.NotImplementedError, state, op, f"Cannot compare types using: {op}")
+        raise exc.Signal.make(T.NotImplementedError, op, f"Cannot compare types using: {op}")
     return new_value_instance(a == b)
 
 @dp_inst
-def _compare(state, op, a: T.number, b: T.row):
+def _compare(state, op, a: T.primitive, b: T.row):
     return _compare(state, op, a, b.primary_key())
 
 @dp_inst
-def _compare(state, op, a: T.row, b: T.number):
+def _compare(state, op, a: T.row, b: T.primitive):
     return _compare(state, op, b, a)
 
 @dp_inst
@@ -465,13 +451,13 @@ def compile_to_inst(state: State, cmp: ast.Compare):
     insts = evaluate(state, cmp.args)
 
     if cmp.op == 'in' or cmp.op == '!in':
-        return _contains(state, cmp.op, insts[0], insts[1])
+        return contains(state, cmp.op, insts[0], insts[1])
 
     op = {
         '==': '=',
         '<>': '!=',
     }.get(cmp.op, cmp.op)
-    return _compare(state, op, insts[0], insts[1])
+    return compare(state, op, insts[0], insts[1])
 
 @dy
 def compile_to_inst(state: State, neg: ast.Neg):
@@ -484,24 +470,16 @@ def compile_to_inst(state: State, neg: ast.Neg):
 @dy
 def compile_to_inst(state: State, arith: ast.BinOp):
     args = cast_to_instance(state, arith.args)
-    return _compile_arith(state, arith, *args)
+    return compile_arith(state, arith, *args)
 
 
+def compile_arith(state, op, a, b):
+    res = _compile_arith(state, op, unvectorized(a), unvectorized(b))
+    return objects.inherit_vectorized(res, [a, b])
+
 @dp_inst
 def _compile_arith(state, arith, a: T.any, b: T.any):
-    raise Signal.make(T.TypeError, state, arith.op, f"Operator '{arith.op}' not implemented for {a.type} and {b.type}")
-
-@dp_inst
-def _compile_arith(state, arith, a: T.vectorized, b: T.vectorized):
-    return vectorized(_compile_arith(state, arith, unvectorized(a), unvectorized(b)))
-
-@dp_inst
-def _compile_arith(state, arith, a: T.any, b: T.vectorized):
-    return vectorized(_compile_arith(state, arith, a, unvectorized(b)))
-
-@dp_inst
-def _compile_arith(state, arith, a: T.vectorized, b: T.any):
-    return vectorized(_compile_arith(state, arith, unvectorized(a), b))
+    raise Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not implemented for {a.type} and {b.type}")
 
 
 
@@ -518,7 +496,7 @@ def _compile_arith(state, arith, a: T.collection, b: T.collection):
     try:
         op = ops[arith.op]
     except KeyError:
-        raise Signal.make(T.TypeError, state, arith.op, f"Operation '{arith.op}' not supported for tables ({a.type}, {b.type})")
+        raise Signal.make(T.TypeError, arith.op, f"Operation '{arith.op}' not supported for tables ({a.type}, {b.type})")
 
     return state.get_var(op).func(state, a, b)
 
@@ -539,9 +517,10 @@ def _compile_arith(state, arith, a: T.primitive, b: T.aggregate):
 @dp_inst
 def _compile_arith(state, arith, a: T.string, b: T.int):
     if arith.op != '*':
-        raise Signal.make(T.TypeError, state, arith.op, f"Operator '{arith.op}' not supported between string and integer.")
+        raise Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not supported between string and integer.")
     return call_pql_func(state, "repeat", [a, b])
 
+
 @dp_inst
 def _compile_arith(state, arith, a: T.number, b: T.number):
     if arith.op in ('/', '**') or a.type <= T.float or b.type <= T.float:
@@ -560,7 +539,7 @@ def _compile_arith(state, arith, a: T.number, b: T.number):
             '**': operator.pow,
         }[arith.op]
     except KeyError:
-        raise Signal.make(T.TypeError, state, arith, f"Operator {arith.op} not supported between types '{a.type}' and '{b.type}'")
+        raise Signal.make(T.TypeError, arith, f"Operator {arith.op} not supported between types '{a.type}' and '{b.type}'")
 
     if settings.optimize and isinstance(a, objects.ValueInstance) and isinstance(b, objects.ValueInstance):
         # Local folding for better performance.
@@ -569,7 +548,7 @@ def _compile_arith(state, arith, a: T.number, b: T.number):
         try:
             value = f(a.local_value, b.local_value)
         except ZeroDivisionError as e:
-            raise Signal.make(T.ValueError, state, arith.args[-1], str(e))
+            raise Signal.make(T.ValueError, arith.args[-1], str(e))
         if arith.op == '**':
             value = float(value)
         return new_value_instance(value, res_type)
@@ -584,7 +563,7 @@ def _compile_arith(state, arith, a: T.string, b: T.string):
         return objects.Instance.make(code, T.bool, [a, b])
 
     if arith.op != '+':
-        raise exc.Signal.make(T.TypeError, state, arith.op, f"Operator '{arith.op}' not supported for strings.")
+        raise exc.Signal.make(T.TypeError, arith.op, f"Operator '{arith.op}' not supported for strings.")
 
     if settings.optimize and isinstance(a, objects.ValueInstance) and isinstance(b, objects.ValueInstance):
         # Local folding for better performance (optional, for better performance)
@@ -597,7 +576,7 @@ def _compile_arith(state, arith, a: T.string, b: T.string):
 
 @dy
 def compile_to_inst(state: State, x: ast.Ellipsis):
-    raise Signal.make(T.SyntaxError, state, x, "Ellipsis not allowed here")
+    raise Signal.make(T.SyntaxError, x, "Ellipsis not allowed here")
 
 
 @dy
@@ -621,7 +600,7 @@ def compile_to_inst(state: State, lst: objects.PythonList):
     x = [sql.Primitive(t, sql._repr(t,i)) for i in (lst.items)]
     name = state.unique_name("list_")
     table_code, subq, list_type = sql.create_list(name, x)
-    inst = objects.ListInstance.make(table_code, list_type, [])
+    inst = objects.TableInstance.make(table_code, list_type, [])
     inst.subqueries[name] = subq
     return inst
 
@@ -637,19 +616,25 @@ def compile_to_inst(state: State, lst: ast.List_):
     types = {e.type for e in elems}
 
     if len(types) > 1:
-        raise Signal.make(T.TypeError, state, lst, f"List members must be of the same type. Got {types}")
+        raise Signal.make(T.TypeError, lst, f"List members must be of the same type. Got {types}")
 
     elem_type ,= types
 
-    if not (elem_type <= T.union[T.primitive, T.nulltype]):
-        raise Signal.make(T.TypeError, state, lst, "Cannot create lists of type %s" % elem_type)
+    if elem_type <= T.struct:
+        rows = [sql.ValuesTuple(obj.type, obj.flatten_code()) for obj in elems]
+        list_type = T.table(elems=elem_type.elems)
+        name = state.unique_name("table_")
+        table_code, subq = sql.create_table(list_type, name, rows)
+    else:
+        if not (elem_type <= T.union[T.primitive, T.nulltype]):
+            raise Signal.make(T.TypeError, lst, "Cannot create lists of type %s" % elem_type)
 
-    assert elem_type <= lst.type.elems[ITEM_NAME], (elem_type, lst.type)
+        assert elem_type <= lst.type.elems[ITEM_NAME], (elem_type, lst.type)
 
-    name = state.unique_name("list_")
-    table_code, subq, list_type = sql.create_list(name, [e.code for e in elems])
+        name = state.unique_name("list_")
+        table_code, subq, list_type = sql.create_list(name, [e.code for e in elems])
 
-    inst = objects.ListInstance.make(table_code, list_type, elems)
+    inst = objects.TableInstance.make(table_code, list_type, elems)
     inst.subqueries[name] = subq
     return inst
 
@@ -693,7 +678,7 @@ def _resolve_sql_parameters(state, compiled_sql, wrap=False, subqueries=None):
         if isinstance(c, sql.Parameter):
             inst = evaluate(state, state.get_var(c.name))
             if inst.type != c.type:
-                raise Signal.make(T.CastError, state, None, f"Internal error: Parameter is of wrong type ({c.type} != {inst.type})")
+                raise Signal.make(T.CastError, None, f"Internal error: Parameter is of wrong type ({c.type} != {inst.type})")
             new_code += inst.code.compile_wrap(qb).code
             subqueries.update(inst.subqueries)
         else:
@@ -711,7 +696,7 @@ def _resolve_sql_parameters(state, compiled_sql, wrap=False, subqueries=None):
 def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
     sql_code = cast_to_python(state, rps.string)
     if not isinstance(sql_code, str):
-        raise Signal.make(T.TypeError, state, rps, f"Expected string, got '{rps.string}'")
+        raise Signal.make(T.TypeError, rps, f"Expected string, got '{rps.string}'")
 
     type_ = evaluate(state, rps.type)
     if isinstance(type_, objects.Instance):
@@ -723,7 +708,6 @@ def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
     else:
         self_table = None
 
-    params_vectorized = False
     instances = []
     subqueries = SafeDict()
     tokens = re_split(r"\$\w+", sql_code)
@@ -733,7 +717,7 @@ def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
             assert t[0] == '$'
             if t == '$self':
                 if self_table is None:
-                    raise exc.Signal.make(T.TypeError, state, rps, f"$self is only available for queries that return a table")
+                    raise exc.Signal.make(T.TypeError, rps, f"$self is only available for queries that return a table")
                 inst = self_table
             else:
                 obj = state.get_var(t[1:])
@@ -743,8 +727,6 @@ def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
                     inst = objects.new_table(obj)
                 else:
                     inst = cast_to_instance(state, obj)
-                    if inst.type <= T.vectorized:
-                        params_vectorized = True
 
             instances.append(inst)
             new_code += _resolve_sql_parameters(state, inst.code, wrap=bool(new_code), subqueries=subqueries).code
@@ -765,8 +747,7 @@ def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
         inst.subqueries[name] = subq
         return inst
 
-    if params_vectorized:
-        type_ = T.vectorized[type_]
+    type_ = inherit_vectorized_type(type_, instances)
     code = sql.CompiledSQL(type_, new_code, None, False, False)     # XXX is False correct?
     return make_instance(code, type_, instances)
 
@@ -774,7 +755,7 @@ def compile_to_inst(state: State, rps: ast.ParameterizedSqlCode):
 def compile_to_inst(state: State, s: ast.Slice):
     obj = cast_to_instance(state, s.obj)
 
-    assert_type(obj.type, T.union[T.string, T.collection, T.vectorized[T.string]], state, s, "Slice")
+    assert_type(kernel_type(obj.type), T.union[T.string, T.collection], state, s, "Slice")
 
     instances = [obj]
     if s.range.start:
@@ -789,7 +770,7 @@ def compile_to_inst(state: State, s: ast.Slice):
     else:
         stop = None
 
-    if obj.type <= T.string or obj.type <= T.vectorized[T.string]:
+    if kernel_type(obj.type) <= T.string:
         code = sql.StringSlice(obj.code, sql.add_one(start.code), stop and sql.add_one(stop.code))
     else:
         start_n = cast_to_python(state, start)
@@ -806,9 +787,9 @@ def compile_to_inst(state: State, sel: ast.Selection):
 
     table = cast_to_instance(state, obj)
 
-    if table.type <= T.string or table.type <= T.vectorized[T.string]:
+    if kernel_type(table.type) <= T.string:
         index ,= cast_to_instance(state, sel.conds)
-        assert index.type <= T.int
+        assert kernel_type(index.type) <= T.int, index.type
         table = table.replace(type=T.string)    # XXX why get rid of vectorized here? because it's a table operation node?
         slice = ast.Slice(table, ast.Range(index, ast.BinOp('+', [index, ast.Const(T.int, 1)]))).set_text_ref(sel.text_ref)
         return compile_to_inst(state, slice)
@@ -822,8 +803,8 @@ def compile_to_inst(state: State, sel: ast.Selection):
         code = sql.unknown
     else:
         for i, c in enumerate(conds):
-            if not (c.type <= T.union[T.bool, T.vectorized[T.bool]]):
-                raise exc.Signal.make(T.TypeError, state, sel.conds[i], f"Selection expected boolean, got {c.type}")
+            if not (kernel_type(c.type) <= T.bool):
+                raise exc.Signal.make(T.TypeError, sel.conds[i], f"Selection expected boolean, got {c.type}")
 
         code = sql.table_selection(table, [c.code for c in conds])
 
@@ -850,38 +831,38 @@ def compile_to_inst(state: State, attr: ast.Attr):
         raise AutocompleteSuggestions(attrs)
 
     if not attr.expr:
-        raise Signal.make(T.NotImplementedError, state, attr, "Implicit attribute syntax not supported")
+        raise Signal.make(T.NotImplementedError, attr, "Implicit attribute syntax not supported")
 
     inst = evaluate(state, attr.expr)
     try:
         return evaluate(state, inst.get_attr(attr.name))
     except exc.pql_AttributeError as e:
-        raise Signal.make(T.AttributeError, state, attr, e.message)
+        raise Signal.make(T.AttributeError, attr, e.message)
 
 
 
 def _apply_type_generics(state, gen_type, type_names):
     type_objs = evaluate(state, type_names)
     if not type_objs:
-        raise Signal.make(T.TypeError, state, None, f"Generics expression expected a type, got nothing.")
+        raise Signal.make(T.TypeError, None, f"Generics expression expected a type, got nothing.")
     for o in type_objs:
         if not isinstance(o, Type):
             if isinstance(o.code, sql.Parameter):
                 # XXX hacky test, hacky solution
                 raise exc.InsufficientAccessLevel()
-            raise Signal.make(T.TypeError, state, None, f"Generics expression expected a type, got '{o}'.")
+            raise Signal.make(T.TypeError, None, f"Generics expression expected a type, got '{o}'.")
 
     if len(type_objs) > 1:
         if gen_type in (T.union,):
             return gen_type(tuple(type_objs))
 
-        raise Signal.make(T.TypeError, state, None, "Union types not yet supported!")
+        raise Signal.make(T.TypeError, None, "Union types not yet supported!")
 
     t ,= type_objs
     try:
         return gen_type[t]
     except TypeError:
-        raise Signal.make(T.TypeError, state, None, f"Type {t} isn't a container!")
+        raise Signal.make(T.TypeError, None, f"Type {t} isn't a container!")
 
 
 
@@ -912,23 +893,39 @@ def compile_to_inst(state: State, marker: ast.Marker):
 
 @dy
 def compile_to_inst(state: State, range: ast.Range):
+    # TODO move to sql.py
+    # Requires subqueries to be part of 'code' instead of a separate 'subqueries'?
+    # But then what's the point of an instance, other than carrying methods...
     start = cast_to_python(state, range.start) if range.start else 0
     if not isinstance(start, int):
-        raise Signal.make(T.TypeError, state, range, "Range must be between integers")
+        raise Signal.make(T.TypeError, range, "Range must be between integers")
+
+    stop = None
     if range.stop:
         stop = cast_to_python(state, range.stop)
         if not isinstance(stop, int):
-            raise Signal.make(T.TypeError, state, range, "Range must be between integers")
-        stop_str = f" WHERE item+1<{stop}"
-    else:
-        if state.db.target is sql.mysql:
-            raise Signal.make(T.NotImplementedError, state, range, "MySQL doesn't support infinite recursion!")
-        stop_str = ''
+            raise Signal.make(T.TypeError, range, "Range must be between integers")
+    elif state.db.target in (sql.mysql, sql.bigquery):
+            raise Signal.make(T.NotImplementedError, range, f"{state.db.target} doesn't support infinite series!")
 
     type_ = T.list[T.int]
+
+    if state.db.target == sql.bigquery:
+        code = sql.RawSql(type_, f'UNNEST(GENERATE_ARRAY({start}, {stop-1})) as item')
+        return objects.TableInstance.make(code, type_, [])
+
+    if stop is None:
+        stop_str = ''
+    else:
+        stop_str = f" WHERE item+1<{stop}"
+
+
     name = state.unique_name("range")
     skip = 1
     code = f"SELECT {start} AS item UNION ALL SELECT item+{skip} FROM {name}{stop_str}"
     subq = sql.Subquery(name, [], sql.RawSql(type_, code))
     code = sql.TableName(type_, Id(name))
-    return objects.ListInstance(code, type_, SafeDict({name: subq}))
+    return objects.TableInstance(code, type_, SafeDict({name: subq}))
+
+
+    
\ No newline at end of file
