@@ -15,8 +15,7 @@ from .parser import Str
 from .interp_common import State, dy, new_value_instance, cast_to_python
 from .compiler import compile_to_inst, cast_to_instance
 from .pql_types import T, Type, Object, Id
-from .types_impl import table_params, table_flat_for_insert, flatten_type, pql_repr
-from .pql_objects import vectorized
+from .types_impl import table_params, table_flat_for_insert, flatten_type, pql_repr, kernel_type
 from .display import display
 
 
@@ -31,6 +30,7 @@ def resolve(state: State, struct_def: ast.StructDef):
 def resolve(state: State, table_def: ast.TableDef):
     name = table_def.name
     if is_global_scope(state):
+        name = state.db.qualified_name(name)
         temporary = False
     else:
         name = '__local_' + state.unique_name(name)
@@ -59,8 +59,7 @@ def resolve(state: State, col_def: ast.ColumnDef):
         return T.t_relation[coltype.type](rel={'table': coltype.parent.type, 'column': coltype.name, 'key': False}).replace(_nullable=coltype.type._nullable)
 
     elif coltype <= T.table:
-        # TODO what if 'id' isn't an int?
-        return T.t_relation[T.int.as_nullable()](rel={'table': coltype, 'column': 'id', 'key': True}).replace(_nullable=coltype._nullable)
+        return T.t_relation[T.t_id.as_nullable()](rel={'table': coltype, 'column': 'id', 'key': True}).replace(_nullable=coltype._nullable)
 
     return coltype(default=col_def.default)
 
@@ -86,7 +85,7 @@ def db_query(state: State, sql_code, subqueries=None):
     try:
         return state.db.query(sql_code, subqueries, state=state)
     except exc.DatabaseQueryError as e:
-        raise Signal.make(T.DbQueryError, state, None, e.args[0]) from e
+        raise Signal.make(T.DbQueryError, None, e.args[0]) from e
 
 def drop_table(state, table_type):
     name ,= table_type.options['name'].parts
@@ -115,15 +114,16 @@ def _execute(state: State, table_def: ast.TableDef):
 
     if any(isinstance(c, ast.Ellipsis) for c in table_def.columns):
         # XXX why must it? just ensure it appears once
-        raise Signal.make(T.SyntaxError, state, table_def, "Ellipsis must appear at the end")
+        raise Signal.make(T.SyntaxError, table_def, "Ellipsis must appear at the end")
 
     # Create type and a corresponding table in the database
     t = resolve(state, table_def)
+    db_name = t.options['name']
 
-    # exists = table_exists(state, table_def.name)
-    exists = state.db.table_exists(table_def.name)
+    exists = state.db.table_exists(db_name.repr_name)
     if exists:
-        cur_type = state.db.import_table_type(state, table_def.name, None if ellipsis else set(t.elems) | {'id'})
+        assert not t.options['temporary']
+        cur_type = state.db.import_table_type(state, db_name.repr_name, None if ellipsis else set(t.elems) | {'id'})
 
         if ellipsis:
             elems_to_add = {Str(n, ellipsis.text_ref): v for n, v in cur_type.elems.items() if n not in t.elems}
@@ -133,32 +133,32 @@ def _execute(state: State, table_def: ast.TableDef):
         # Auto-add id only if it exists already and not defined by user
         if 'id' in cur_type.elems: #and 'id' not in t.elems:
             # t = t(dict(id=T.t_id, **t.elems), pk=[['id']])
-            assert cur_type.elems['id'] <= T.int
+            assert cur_type.elems['id'] <= T.primitive, cur_type.elems['id']
             t.elems['id'] = T.t_id
 
         for e_name, e1_type in t.elems.items():
 
             if e_name not in cur_type.elems:
-                raise Signal.make(T.TypeError, state, table_def, f"Column '{e_name}' defined, but doesn't exist in database.")
+                raise Signal.make(T.TypeError, table_def, f"Column '{e_name}' defined, but doesn't exist in database.")
 
-            e2_type = cur_type.elems[e_name]
+            # e2_type = cur_type.elems[e_name]
             # XXX use can_cast() instead of hardcoding it
             # if not (e1_type <= e2_type or (e1_type <= T.t_id and e2_type <= T.int)):
-            #     raise Signal.make(T.TypeError, state, table_def, f"Cannot cast column '{e_name}' from type '{e2_type}' to '{e1_type}'")
+            #     raise Signal.make(T.TypeError, table_def, f"Cannot cast column '{e_name}' from type '{e2_type}' to '{e1_type}'")
 
-        inst = objects.new_table(t, Id(table_def.name), select_fields=True)
+        inst = objects.new_table(t, db_name, select_fields=True)
     else:
         # Auto-add id by default
         elems = dict(t.elems)
         if 'id' not in elems:
             elems = {'id': T.t_id, **elems}
         t = t(elems, pk=[['id']])
-        inst = objects.new_table(t, Id(table_def.name))
+        inst = objects.new_table(t, db_name)
 
     state.set_var(table_def.name, inst)
 
     if not exists:
-        sql_code = sql.compile_type_def(state, table_def.name, t)
+        sql_code = sql.compile_type_def(state, db_name.repr_name, t)
         db_query(state, sql_code)
 
 @dy
@@ -167,7 +167,7 @@ def _set_value(state: State, name: ast.Name, value):
 
 @dy
 def _set_value(state: State, attr: ast.Attr, value):
-    raise Signal.make(T.NotImplementedError, state, attr, f"Cannot set attribute for {attr.expr.repr(state)}")
+    raise Signal.make(T.NotImplementedError, attr, f"Cannot set attribute for {attr.expr.repr()}")
 
 @dy
 def _execute(state: State, var_def: ast.SetValue):
@@ -187,7 +187,7 @@ def _copy_rows(state: State, target_name: ast.Name, source: objects.TableInstanc
     params = dict(table_params(target.type))
     for p in params:
         if p not in source.type.elems:
-            raise Signal.make(T.TypeError, state, source, f"Missing column '{p}' in {source.type}")
+            raise Signal.make(T.TypeError, source, f"Missing column '{p}' in {source.type}")
 
     primary_keys, columns = table_flat_for_insert(target.type)
 
@@ -234,7 +234,7 @@ def _execute(state: State, p: ast.Print):
         if inst.type <= T.string:
             repr_ = cast_to_python(state, inst)
         else:
-            repr_ = inst.repr(state)
+            repr_ = inst.repr()
 
         display.print(repr_, end=" ")
     display.print("")
@@ -245,10 +245,10 @@ def _execute(state: State, p: ast.Assert):
     if not res:
         # TODO pretty print values
         if isinstance(p.cond, ast.Compare):
-            s = (' %s '%p.cond.op).join(str(evaluate(state, a).repr(state)) for a in p.cond.args)
+            s = (' %s '%p.cond.op).join(str(evaluate(state, a).repr()) for a in p.cond.args)
         else:
-            s = p.cond.repr(state)
-        raise Signal.make(T.AssertError, state, p.cond, f"Assertion failed: {s}")
+            s = p.cond.repr()
+        raise Signal.make(T.AssertError, p.cond, f"Assertion failed: {s}")
 
 @dy
 def _execute(state: State, cb: ast.CodeBlock):
@@ -294,11 +294,9 @@ def import_module(state, r):
     for path in paths:
         module_path =  (path / r.module_path).with_suffix(".pql")
         if module_path.exists():
-            # with open(module_path, encoding='utf8') as f:
-            #     text = f.read()
             break
     else:
-        raise Signal.make(T.ImportError, state, r, "Cannot find module")
+        raise Signal.make(T.ImportError, r, "Cannot find module")
 
     from .interpreter import Interpreter    # XXX state.new_interp() ?
     i = Interpreter(state.db, state.fmt, use_core=r.use_core)
@@ -414,9 +412,10 @@ def simplify(state: State, x):
 @dy
 def simplify(state: State, obj: ast.Or):
     a, b = evaluate(state, obj.args)
-    _, (ta, tb) = objects.unvectorize_args([a,b])
-    if ta.type != tb.type:
-        raise Signal.make(T.TypeError, state, obj, f"'or' operator requires both arguments to be of the same type, but got '{ta.type}' and '{tb.type}'.")
+    ta = kernel_type(a.type)
+    tb = kernel_type(b.type)
+    if ta != tb:
+        raise Signal.make(T.TypeError, obj, f"'or' operator requires both arguments to be of the same type, but got '{ta}' and '{tb}'.")
     try:
         if test_nonzero(state, a):
             return a
@@ -428,9 +427,10 @@ def simplify(state: State, obj: ast.Or):
 @dy
 def simplify(state: State, obj: ast.And):
     a, b = evaluate(state, obj.args)
-    _, (ta, tb) = objects.unvectorize_args([a,b])
-    if ta.type != tb.type:
-        raise Signal.make(T.TypeError, state, obj, f"'or' operator requires both arguments to be of the same type, but got '{ta.type}' and '{tb.type}'.")
+    ta = kernel_type(a.type)
+    tb = kernel_type(b.type)
+    if ta != tb:
+        raise Signal.make(T.TypeError, obj, f"'and' operator requires both arguments to be of the same type, but got '{ta}' and '{tb}'.")
     try:
         if not test_nonzero(state, a):
             return a
@@ -456,7 +456,7 @@ def simplify(state: State, funccall: ast.FuncCall):
 
     if isinstance(func, objects.UnknownInstance):
         # evaluate(state, [a.value for a in funccall.args])
-        raise Signal.make(T.TypeError, state, funccall.func, f"Error: Object of type '{func.type}' is not callable")
+        raise Signal.make(T.TypeError, funccall.func, f"Error: Object of type '{func.type}' is not callable")
 
     args = funccall.args
     if isinstance(func, Type):
@@ -465,7 +465,7 @@ def simplify(state: State, funccall: ast.FuncCall):
         func = state.get_var('cast')
 
     if not isinstance(func, objects.Function):
-        raise Signal.make(T.TypeError, state, funccall.func, f"Error: Object of type '{func.type}' is not callable")
+        raise Signal.make(T.TypeError, funccall.func, f"Error: Object of type '{func.type}' is not callable")
 
     state.stacktrace.append(funccall.text_ref)
     try:
@@ -496,9 +496,8 @@ def eval_func_call(state, func, args):
     for i, (p, a) in enumerate(matched_args):
         a = evaluate(state, a)
         # TODO cast?
-        # if p.type and not a.type <= T.union[p.type, T.vectorized[p.type]]:
-        if p.type and not a.type <= p.type:
-            raise Signal.make(T.TypeError, state, func, f"Argument #{i} of '{func.name}' is of type '{a.type}', expected '{p.type}'")
+        if p.type and not kernel_type(a.type) <= p.type:
+            raise Signal.make(T.TypeError, func, f"Argument #{i} of '{func.name}' is of type '{a.type}', expected '{p.type}'")
         args[p.name] = a
 
 
@@ -506,12 +505,7 @@ def eval_func_call(state, func, args):
         # TODO ensure pure function?
         # TODO Ensure correct types
         args = list(args.values())
-        # args = evaluate(state, args)
-        # was_vec, args = objects.unvectorize_args(args)
-        res = func.func(state, *args)
-        # if was_vec:
-        #     res = vectorized(res)
-        return res
+        return func.func(state, *args)
     else:
         # TODO make tests to ensure caching was successful
         if settings.cache:
@@ -584,7 +578,7 @@ def apply_database_rw(state: State, o: ast.One):
     obj = evaluate(state, o.expr)
     if obj.type <= T.struct:
         if len(obj.attrs) != 1:
-            raise Signal.make(T.ValueError, state, o, f"'one' expected a struct with a single attribute, got {len(obj.attrs)}")
+            raise Signal.make(T.ValueError, o, f"'one' expected a struct with a single attribute, got {len(obj.attrs)}")
         x ,= obj.attrs.values()
         return x
 
@@ -595,10 +589,10 @@ def apply_database_rw(state: State, o: ast.One):
     rows = localize(state, table) # Must be 1 row
     if len(rows) == 0:
         if not o.nullable:
-            raise Signal.make(T.ValueError, state, o, "'one' expected a single result, got an empty expression")
+            raise Signal.make(T.ValueError, o, "'one' expected a single result, got an empty expression")
         return objects.null
     elif len(rows) > 1:
-        raise Signal.make(T.ValueError, state, o, "'one' expected a single result, got more")
+        raise Signal.make(T.ValueError, o, "'one' expected a single result, got more")
 
     row ,= rows
     rowtype = T.row[table.type]
@@ -621,15 +615,15 @@ def apply_database_rw(state: State, d: ast.Delete):
     table = evaluate(state, cond_table)
 
     if not (table.type <= T.table):
-        raise Signal.make(T.TypeError, state, d.table, f"Expected a table. Got: {table.type}")
+        raise Signal.make(T.TypeError, d.table, f"Expected a table. Got: {table.type}")
 
     if not 'name' in table.type.options:
-        raise Signal.make(T.ValueError, state, d.table, "Cannot delete. Table is not persistent")
+        raise Signal.make(T.ValueError, d.table, "Cannot delete. Table is not persistent")
 
     rows = list(localize(state, table))
     if rows:
         if 'id' not in rows[0]:
-            raise Signal.make(T.TypeError, state, d, "Delete error: Table does not contain id")
+            raise Signal.make(T.TypeError, d, "Delete error: Table does not contain id")
 
         ids = [row['id'] for row in rows]
 
@@ -646,14 +640,14 @@ def apply_database_rw(state: State, u: ast.Update):
     table = evaluate(state, u.table)
 
     if not (table.type <= T.table):
-        raise Signal.make(T.TypeError, state, u.table, f"Expected a table. Got: {table.type}")
+        raise Signal.make(T.TypeError, u.table, f"Expected a table. Got: {table.type}")
 
     if not 'name' in table.type.options:
-        raise Signal.make(T.ValueError, state, u.table, "Cannot update: Table is not persistent")
+        raise Signal.make(T.ValueError, u.table, "Cannot update: Table is not persistent")
 
     for f in u.fields:
         if not f.name:
-            raise Signal.make(T.SyntaxError, state, f, f"Update requires that all fields have a name")
+            raise Signal.make(T.SyntaxError, f, f"Update requires that all fields have a name")
 
     # TODO verify table is concrete (i.e. lvalue, not a transitory expression)
 
@@ -664,9 +658,9 @@ def apply_database_rw(state: State, u: ast.Update):
     rows = list(localize(state, table))
     if rows:
         if 'id' not in rows[0]:
-            raise Signal.make(T.TypeError, state, u, "Update error: Table does not contain id")
+            raise Signal.make(T.TypeError, u, "Update error: Table does not contain id")
         if not set(proj) < set(rows[0]):
-            raise Signal.make(T.TypeError, state, u, "Update error: Not all keys exist in table")
+            raise Signal.make(T.TypeError, u, "Update error: Not all keys exist in table")
 
         ids = [row['id'] for row in rows]
 
@@ -684,7 +678,7 @@ def apply_database_rw(state: State, new: ast.NewRows):
     obj = state.get_var(new.type)
 
     if len(new.args) > 1:
-        raise Signal.make(T.NotImplementedError, state, new, "Not yet implemented") #. Requires column-wise table concat (use join and enum)")
+        raise Signal.make(T.NotImplementedError, new, "Not yet implemented") #. Requires column-wise table concat (use join and enum)")
 
     if isinstance(obj, objects.UnknownInstance):
         arg ,= new.args
@@ -729,9 +723,9 @@ def _destructure_param_match(state, ast_node, param_match):
         if (k.type <= T.struct):
             names = [name for name, t in flatten_type(k.orig, [k.name])]
             if not isinstance(v, list):
-                raise Signal.make(T.TypeError, state, ast_node, f"Parameter {k.name} received a bad value: {v} (expecting a struct or a list)")
+                raise Signal.make(T.TypeError, ast_node, f"Parameter {k.name} received a bad value: {v} (expecting a struct or a list)")
             if len(v) != len(names):
-                raise Signal.make(T.TypeError, state, ast_node, f"Parameter {k.name} received a bad value (size of {len(names)})")
+                raise Signal.make(T.TypeError, ast_node, f"Parameter {k.name} received a bad value (size of {len(names)})")
             yield from safezip(names, v)
         else:
             yield k.name, v
@@ -758,9 +752,18 @@ def _new_row(state, new_ast, table, matched):
     values = [sql.make_value(v) for (_,v) in destructured_pairs]
     assert keys and values
     # XXX use regular insert?
-    q = sql.InsertConsts(new_ast.type, keys, [values])
-    db_query(state, q)
-    rowid = db_query(state, sql.LastRowId())
+
+    if state.db.target == sql.bigquery:
+        rowid = db_query(state, sql.FuncCall(T.string, 'GENERATE_UUID', []))
+        keys += ['id']
+        values += [sql.make_value(rowid)]
+        q = sql.InsertConsts(table.options['name'].repr_name, keys, [values])
+        db_query(state, q)
+    else:
+        q = sql.InsertConsts(table.options['name'].repr_name, keys, [values])
+        # q = sql.InsertConsts(new_ast.type, keys, [values])
+        db_query(state, q)
+        rowid = db_query(state, sql.LastRowId())
 
     d = SafeDict({'id': objects.new_value_instance(rowid)})
     d.update({p.name:v for p, v in matched})
@@ -785,7 +788,7 @@ def apply_database_rw(state: State, new: ast.New):
         return res
 
     if not isinstance(obj, objects.TableInstance):
-        raise Signal.make(T.TypeError, state, new, f"'new' expects a table or exception, instead got {obj.repr(state)}")
+        raise Signal.make(T.TypeError, new, f"'new' expects a table or exception, instead got {obj.repr()}")
 
     table = obj
     # TODO assert tabletype is a real table and not a query (not transient), otherwise new is meaningless
@@ -909,36 +912,6 @@ def localize(state, x):
 
 
 
-### Added functions
-
-def function_help_str(self, state):
-    params = []
-    for p in self.params:
-        s = p.name
-        if p.type:
-            s += f": {p.type}"
-        if p.default:
-            s += f"={p.default.repr(state)}"
-        params.append(s)
-
-    if self.param_collector is not None:
-        params.append(f"...{self.param_collector.name}")
-    param_str = ', '.join(params)
-    return f"func {self.name}({param_str}) = ..."
-
-def function_localize_keys(self, state, struct):
-    return cast_to_python(state, struct)
-
-objects.Function.help_str = function_help_str
-objects.Function._localize_keys = function_localize_keys
-
-
-def instance_repr(self, state):
-    return pql_repr(state, self.type, localize(state, self))
-
-objects.Instance.repr = instance_repr
-
-
 
 
 def new_table_from_rows(state, name, columns, rows):
@@ -961,6 +934,7 @@ def new_table_from_rows(state, name, columns, rows):
 
     x = objects.new_table(table)
     state.set_var(name, x)
+    return x
 
 
 def new_table_from_expr(state, name, expr, const, temporary):
@@ -970,7 +944,7 @@ def new_table_from_expr(state, name, expr, const, temporary):
         return objects.TableInstance.make(sql.null, expr.type, [])
 
     if 'id' in elems and not const:
-        raise Signal.make(T.NameError, state, None, "Field 'id' already exists. Rename it, or use 'const table' to copy it as-is.")
+        raise Signal.make(T.NameError, None, "Field 'id' already exists. Rename it, or use 'const table' to copy it as-is.")
 
     table = T.table(dict(elems), name=Id(name), pk=[] if const else [['id']], temporary=temporary)
 
@@ -988,7 +962,7 @@ def new_table_from_expr(state, name, expr, const, temporary):
 
 @dy
 def cast_to_python(state, obj):
-    raise Signal.make(T.TypeError, state, None, f"Unexpected value: {pql_repr(state, obj.type, obj)}")
+    raise Signal.make(T.TypeError, None, f"Unexpected value: {pql_repr(obj.type, obj)}")
 
 @dy
 def cast_to_python(state, obj: ast.Ast):
@@ -1000,7 +974,7 @@ def cast_to_python(state, obj: objects.AbsInstance):
     # if state.access_level <= state.AccessLevels.QUERY:
     if obj.type <= T.vectorized:
         raise exc.InsufficientAccessLevel(state.access_level)
-        # raise Signal.make(T.CastError, state, None, f"Internal error. Cannot cast vectorized (i.e. projected) obj: {obj}")
+        # raise Signal.make(T.CastError, None, f"Internal error. Cannot cast vectorized (i.e. projected) obj: {obj}")
     res = localize(state, obj)
     if obj.type == T.float:
         res = float(res)
@@ -1012,3 +986,19 @@ def cast_to_python(state, obj: objects.AbsInstance):
     assert isinstance(res, (int, str, float, dict, list, type(None))), (res, type(res))
     return res
 
+
+
+### Added functions
+
+def function_localize_keys(self, state, struct):
+    return cast_to_python(state, struct)
+
+objects.Function._localize_keys = function_localize_keys
+
+
+from .context import context
+def instance_repr(self):
+    return pql_repr(self.type, localize(context.state, self))
+
+objects.Instance.repr = instance_repr
+
