@@ -1,12 +1,13 @@
 import inspect
 from preql.compiler import cast_to_instance
 import re
+import os
+from datetime import datetime
 import csv
 import inspect
 import itertools
-from typing import Optional
-
-from tqdm import tqdm
+import runtype
+import json
 
 from .utils import safezip, listgen, re_split
 from .exceptions import Signal, ExitInterp
@@ -16,10 +17,11 @@ from . import pql_ast as ast
 from . import sql
 
 from .interp_common import State, new_value_instance, assert_type
-from .evaluate import evaluate, cast_to_python, db_query, TableConstructor, new_table_from_expr
+from .evaluate import evaluate, cast_to_python, db_query, TableConstructor, new_table_from_expr, new_table_from_rows
 from .pql_types import T, Type, union_types, Id, common_type
 from .types_impl import join_names
 from .casts import cast
+from .docstring.autodoc import autodoc
 
 def new_str(x):
     return new_value_instance(str(x), T.string)
@@ -34,13 +36,25 @@ def _pql_PY_callback(state: State, var: str):
     inst = evaluate(state, obj)
 
     if not isinstance(inst, objects.ValueInstance):
-        raise Signal.make(T.TypeError, state, None, f"Cannot convert {inst} to a Python value")
+        raise Signal.make(T.TypeError, None, f"Cannot convert {inst} to a Python value")
 
     return str(inst.local_value)
 
 
 def pql_PY(state: State, code_expr: T.string, code_setup: T.string.as_nullable() = objects.null):
-    "Run the given Python code and convert the result to a Preql object"
+    """Evaluate the given Python expression and convert the result to a Preql object
+
+    Parameters:
+        code_expr: The Python expression to evaluate
+        code_setup: Setup code to prepare for the evaluation
+
+    Note:
+        This function is still experemental, and should be used with caution.
+
+    Example:
+        >> PY("sys.version", "import sys")
+        "3.8.2 (tags/v3.8.2:7b3ab59, Feb 25 2020, 23:03:10)"
+    """
     py_code = cast_to_python(state, code_expr)
     py_setup = cast_to_python(state, code_setup)
 
@@ -50,37 +64,81 @@ def pql_PY(state: State, code_expr: T.string, code_setup: T.string.as_nullable()
         try:
             res = exec(py_setup)
         except Exception as e:
-            raise Signal.make(T.EvalError, state, code_expr, f"Python code provided returned an error: {e}")
+            raise Signal.make(T.EvalError, code_expr, f"Python code provided returned an error: {e}")
 
     try:
         res = eval(py_code)
     except Exception as e:
-        raise Signal.make(T.EvalError, state, code_expr, f"Python code provided returned an error: {e}")
+        raise Signal.make(T.EvalError, code_expr, f"Python code provided returned an error: {e}")
 
     return objects.from_python(res)
     # return new_value_instance(res)
 
 def pql_inspect_sql(state: State, obj: T.object):
-    "Returns the SQL code that would be executed to evaluate the given object"
+    """Returns the SQL code that would be executed to evaluate the given object
+
+    """
     if not isinstance(obj, objects.Instance):
-        raise Signal.make(T.TypeError, state, None, f"inspect_sql() expects a concrete object. Instead got: {obj.type}")
+        raise Signal.make(T.TypeError, None, f"inspect_sql() expects a concrete object. Instead got: {obj.type}")
     s = state.db.compile_sql(obj.code, obj.subqueries)
     return objects.ValueInstance.make(sql.make_value(s), T.text, [], s)
 
 
 def pql_SQL(state: State, result_type: T.union[T.collection, T.type], sql_code: T.string):
-    """Create an object with the given SQL evaluation code, and given result type
+    """Create an object with the given SQL evaluation code, and given result type.
+
+    The object will only be evaluated when required by the program flow.
+
+    Parameters:
+        result_type: The expected type of the result of the SQL query
+        sql_code: The SQL code to be evaluated
+
+    Example:
+        >> ["a", "b"]{item: SQL(string, "$item || '!'")}
+        table  =2
+        ┏━━━━━━━┓
+        ┃ item  ┃
+        ┡━━━━━━━┩
+        │ a!    │
+        │ b!    │
+        └───────┘
+
+        >> x = ["a", "b", "c"]
+        >> SQL(type(x), "SELECT item || '!' FROM $x")
+        table  =3
+        ┏━━━━━━━┓
+        ┃ item  ┃
+        ┡━━━━━━━┩
+        │ a!    │
+        │ b!    │
+        │ c!    │
+        └───────┘
     """
     # TODO optimize for when the string is known (prefetch the variables and return Sql)
     # .. why not just compile with parameters? the types are already known
     return ast.ParameterizedSqlCode(result_type, sql_code)
 
 def pql_force_eval(state: State, expr: T.object):
-    "Force evaluation of expression. Execute any db queries necessary."
+    """Forces the evaluation of the given expression.
+
+    Executes any db queries necessary.
+    """
     return objects.new_value_instance( cast_to_python(state, expr) )
 
 def pql_fmt(state: State, s: T.string):
-    "Format given string using interpolation on variables marked as `$var`"
+    """Format the given string using interpolation on variables marked as `$var`
+
+    Example:
+        >> ["a", "b", "c"]{item: fmt("$item!")}
+        table  =3
+        ┏━━━━━━━┓
+        ┃ item  ┃
+        ┡━━━━━━━┩
+        │ a!    │
+        │ b!    │
+        │ c!    │
+        └───────┘
+    """
     _s = cast_to_python(state, s)
 
     tokens = re_split(r"\$\w+", _s)
@@ -125,12 +183,18 @@ def pql_brk_continue(state):
 
 
 def pql_breakpoint(state: State):
-    "Create a Python breakpoint. For a Preql breakpoint, use debug() instead"
+    """Creates a Python breakpoint.
+
+    For a Preql breakpoint, use debug() instead
+    """
     breakpoint()
     return objects.null
 
 def pql_debug(state: State):
-    "Hop into a debug session with REPL"
+    """Breaks the execution of the interpreter, and enters into a debug
+    session using the REPL environment.
+
+    """
     py_api = state._py_api
 
     with state.use_scope(breakpoint_funcs):
@@ -139,48 +203,97 @@ def pql_debug(state: State):
 
 
 
-def pql_issubclass(state: State, inst: T.any, type: T.type):
-    "Returns whether the give object is an instance of the given type"
-    type_ = type
-    assert_type(inst.type, T.type, state, inst, 'issubclass')
-    assert_type(type_.type, T.type, state, inst, 'issubclass')
-    assert isinstance(inst, Type)
-    assert isinstance(type_, Type)
-    res = inst <= type_
+def pql_issubclass(state: State, a: T.type, b: T.type):
+    """Checks if type 'a' is a subclass of type 'b'
+
+    Examples:
+        >> issubclass(int, number)
+        true
+        >> issubclass(int, table)
+        false
+        >> issubclass(list, table)
+        true
+    """
+    assert_type(a.type, T.type, state, a, 'issubclass')
+    assert_type(b.type, T.type, state, b, 'issubclass')
+    assert isinstance(a, Type)
+    assert isinstance(b, Type)
+    return new_value_instance(a <= b, T.bool)
+
+def pql_isa(state: State, obj: T.any, type: T.type):
+    """Checks if the give object is an instance of the given type
+
+    Examples:
+        >> isa(1, int)
+        true
+        >> isa(1, string)
+        false
+        >> isa(1.2, number)
+        true
+        >> isa([1], table)
+        true
+    """
+    assert_type(type.type, T.type, state, obj, 'isa')
+    res = obj.isa(type)
     return new_value_instance(res, T.bool)
 
-def pql_isa(state: State, inst: T.any, type: T.type):
-    "Returns whether the give object is an instance of the given type"
-    type_ = type
-    assert_type(type_.type, T.type, state, inst, 'isa')
-    res = inst.isa(type_)
-    return new_value_instance(res, T.bool)
-
-def _count(state, obj, table_func, name):
+def _count(state, obj, table_func, name='count'):
     if obj is objects.null:
         code = sql.FieldFunc(name, sql.AllFields(T.any))
     elif obj.type <= T.table:
         code = table_func(obj.code)
     elif isinstance(obj, objects.StructInstance):
         return new_value_instance(len(obj.attrs))
+
+    elif obj.type <= T.vectorized[T.json_array]:
+        code = sql.JsonLength(obj.code)
     else:
         if not (obj.type <= T.aggregate):
-            raise Signal.make(T.TypeError, state, None, f"Function '{name}' expected an aggregated list, but got '{obj.type}' instead. Did you forget to group?")
+            raise Signal.make(T.TypeError, None, f"Function '{name}' expected an aggregated list, but got '{obj.type}' instead. Did you forget to group?")
 
         obj = obj.primary_key()
-        code = sql.FieldFunc(name, obj.code)
+        code = sql.FieldFunc('count', obj.code)
 
     return objects.Instance.make(code, T.int, [obj])
 
 def pql_count(state: State, obj: T.container.as_nullable() = objects.null):
-    "Count how many rows are in the given table, or in the projected column."
-    return _count(state, obj, sql.CountTable, 'count')
+    """Count how many rows are in the given table, or in the projected column.
+
+    If no argument is given, count all the rows in the current projection.
+
+    Examples:
+        >> count([0..10])
+        10
+        >> [0..10]{ => count() }
+        table  =1
+        ┏━━━━━━━┓
+        ┃ count ┃
+        ┡━━━━━━━┩
+        │    10 │
+        └───────┘
+        >> [0..10]{ => count(item) }
+        table  =1
+        ┏━━━━━━━┓
+        ┃ count ┃
+        ┡━━━━━━━┩
+        │    10 │
+        └───────┘
+    """
+    return _count(state, obj, sql.CountTable)
 
 
 def pql_temptable(state: State, expr: T.collection, const: T.bool.as_nullable() = objects.null):
     """Generate a temporary table with the contents of the given table
 
-    It will remain available until the db-session ends, unless manually removed.
+    It will remain available until the database session ends, unless manually removed.
+
+    Parameters:
+        expr: the table expression to create the table from
+        const: whether the resulting table may be changed or not.
+
+    Note:
+        A non-const table creates its own `id` field.
+        Trying to copy an existing id field into it will cause a collision
     """
     # 'temptable' creates its own counting 'id' field. Copying existing 'id' fields will cause a collision
     # 'const table' doesn't
@@ -195,9 +308,9 @@ def pql_temptable(state: State, expr: T.collection, const: T.bool.as_nullable()
 def pql_get_db_type(state: State):
     """Returns a string representing the type of the active database.
 
-    Possible values are:
-        - "sqlite"
-        - "postgres"
+    Example:
+        >> get_db_type()
+        "sqlite"
     """
     assert state.access_level >= state.AccessLevels.EVALUATE
     return new_value_instance(state.db.target, T.string)
@@ -207,19 +320,19 @@ def pql_get_db_type(state: State):
 def sql_bin_op(state, op, t1, t2, name, additive=False):
 
     if not isinstance(t1, objects.CollectionInstance):
-        raise Signal.make(T.TypeError, state, t1, f"First argument isn't a table, it's a {t1.type}")
+        raise Signal.make(T.TypeError, t1, f"First argument isn't a table, it's a {t1.type}")
     if not isinstance(t2, objects.CollectionInstance):
-        raise Signal.make(T.TypeError, state, t2, f"Second argument isn't a table, it's a {t2.type}")
+        raise Signal.make(T.TypeError, t2, f"Second argument isn't a table, it's a {t2.type}")
 
     # TODO Smarter matching?
     l1 = len(t1.type.elems)
     l2 = len(t2.type.elems)
     if l1 != l2:
-        raise Signal.make(T.TypeError, state, None, f"Cannot {name} tables due to column mismatch (table1 has {l1} columns, table2 has {l2} columns)")
+        raise Signal.make(T.TypeError, None, f"Cannot {name} tables due to column mismatch (table1 has {l1} columns, table2 has {l2} columns)")
 
     for e1, e2 in zip(t1.type.elems.values(), t2.type.elems.values()):
         if not (e2 <= e1):
-            raise Signal.make(T.TypeError, state, None, f"Cannot {name}. Column types don't match: '{e1}' and '{e2}'")
+            raise Signal.make(T.TypeError, None, f"Cannot {name}. Column types don't match: '{e1}' and '{e2}'")
 
     code = sql.TableArith(op, [t1.code, t2.code])
 
@@ -232,7 +345,7 @@ def pql_table_intersect(state: State, t1: T.collection, t2: T.collection):
 def pql_table_substract(state: State, t1: T.collection, t2: T.collection):
     "Substract two tables (except). Used for `t1 - t2`"
     if state.db.target is sql.mysql:
-        raise Signal.make(T.NotImplementedError, state, t1, "MySQL doesn't support EXCEPT (yeah, really!)")
+        raise Signal.make(T.NotImplementedError, t1, "MySQL doesn't support EXCEPT (yeah, really!)")
     return sql_bin_op(state, "EXCEPT", t1, t2, "subtract")
 
 def pql_table_union(state: State, t1: T.collection, t2: T.collection):
@@ -253,7 +366,7 @@ def _get_table(t):
         return t.parent
 
     if not isinstance(t, objects.CollectionInstance):
-        raise Signal.make(T.TypeError, state, None, f"join() arguments must be tables")
+        raise Signal.make(T.TypeError, None, f"join() arguments must be tables")
     return t
 
 def _join2(state, join, a, b):
@@ -261,7 +374,9 @@ def _join2(state, join, a, b):
         return [a, b]
 
     if not ((a.type <= T.collection) and (b.type <= T.collection)):
-        raise Signal.make(T.TypeError, state, None, f"join() arguments must be of same type. Instead got:\n * {a}\n * {b}")
+        a = a.type.repr()
+        b = b.type.repr()
+        raise Signal.make(T.TypeError, None, f"join() arguments must be of same type. Instead got:\n * {a}\n * {b}")
 
     return _auto_join(state, join, a, b)
 
@@ -273,11 +388,11 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
     # Validation and edge cases
     for x in exprs:
         if not isinstance(x, objects.AbsInstance):
-            raise Signal.make(T.TypeError, state, None, f"Unexpected object type: {x}")
+            raise Signal.make(T.TypeError, None, f"Unexpected object type: {x}")
 
     for e in exprs:
         if e is objects.EmptyList:
-            raise Signal.make(T.TypeError, state, None, "Cannot join on an untyped empty list")
+            raise Signal.make(T.TypeError, None, "Cannot join on an untyped empty list")
 
         if isinstance(e, objects.UnknownInstance):
             table_type = T.table({n: T.unknown for n in names})
@@ -306,10 +421,10 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
     if joinall:
         for e in exprs:
             if not isinstance(e, objects.CollectionInstance):
-                raise Signal.make(T.TypeError, state, None, f"joinall() expected tables. Got {e}")
+                raise Signal.make(T.TypeError, None, f"joinall() expected tables. Got {e}")
     else:
         if len(exprs) < 2:
-            raise Signal.make(T.TypeError, state, None, "join expected at least 2 arguments")
+            raise Signal.make(T.TypeError, None, "join expected at least 2 arguments")
 
         joined_exprs = set()
         for (na, ta), (nb, tb) in itertools.combinations(safezip(names, exprs), 2):
@@ -323,7 +438,7 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
         if {id(e) for e in exprs} != set(joined_exprs):
             # TODO better error!!! table name?? specific failed auto-join?
             s = ', '.join(repr(t.type) for t in exprs)
-            raise Signal.make(T.JoinError, state, None, f"Cannot auto-join: No plausible relations found between {s}")
+            raise Signal.make(T.JoinError, None, f"Cannot auto-join: No plausible relations found between {s}")
 
 
     code = sql.Join(table_type, join, [t.code for t in tables], conds)
@@ -331,31 +446,83 @@ def _join(state: State, join: str, exprs_dict: dict, joinall=False, nullable=Non
 
 
 def pql_join(state, tables):
-    "Inner join two tables into a new projection {t1, t2}"
+    """Inner-join any number of tables.
+
+    Each argument is expected to be one of -
+    (1) A column to join on. Columns are attached to specific tables. or
+    (2) A table to join on. The column will be chosen automatically, if there is no ambiguity.
+    Connections are made according to the relationships in the declaration of the table.
+
+    Parameters:
+        tables: Each argument must be either a column, or a table.
+
+    Returns:
+        A new table, where each column is a struct representing one of
+        the joined tables.
+
+    Examples:
+        >> join(a: [0].item, b: [0].item)
+        table join46 =1
+        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┓
+        ┃ a           ┃ b           ┃
+        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━┩
+        │ {'item': 0} │ {'item': 0} │
+        └─────────────┴─────────────┘
+
+        >> join(a: [1..5].item, b: [3..8].item) {...a}
+        table  =2
+        ┏━━━━━━━┓
+        ┃  item ┃
+        ┡━━━━━━━┩
+        │     3 │
+        │     4 │
+        └───────┘
+
+        >> join(c: Country, l: Language) {...c, language: l.name}
+    """
     return _join(state, "JOIN", tables)
 def pql_leftjoin(state, tables):
-    "Left join two tables into a new projection {t1, t2}"
+    """Left-join any number of tables
+
+    See `join`
+    """
     return _join(state, "LEFT JOIN", tables, nullable=[False, True])
 def pql_outerjoin(state, tables):
-    "Outer join two tables into a new projection {t1, t2}"
+    """Outer-join any number of tables
+
+    See `join`
+    """
     return _join(state, "FULL OUTER JOIN", tables, nullable=[False, True])
 def pql_joinall(state: State, tables):
-    "Cartesian product of two tables into a new projection {t1, t2}"
+    """Cartesian product of any number of tables
+
+    See `join`
+
+    Example:
+        >> joinall(a: [0,1], b: ["a", "b"])
+            table joinall14 =4
+        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
+        ┃ a           ┃ b             ┃
+        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
+        │ {'item': 0} │ {'item': 'a'} │
+        │ {'item': 0} │ {'item': 'b'} │
+        │ {'item': 1} │ {'item': 'a'} │
+        │ {'item': 1} │ {'item': 'b'} │
+        └─────────────┴───────────────┘
+    """
     return _join(state, "JOIN", tables, True)
 
 class NoAutoJoinFound(Exception):
     pass
 
 def _auto_join(state, join, ta, tb):
-    # if len(ta.type.elems) == len(tb.type.elems) == 1:
-
     refs1 = _find_table_reference(ta, tb)
     refs2 = _find_table_reference(tb, ta)
     auto_join_count = len(refs1) + len(refs2)
     if auto_join_count < 1:
         raise NoAutoJoinFound(ta, tb)
     elif auto_join_count > 1:   # Ambiguity in auto join resolution
-        raise Signal.make(T.JoinError, state, None, "Cannot auto-join: Several plausible relations found")
+        raise Signal.make(T.JoinError, None, "Cannot auto-join: Several plausible relations found")
 
     if len(refs1) == 1:
         dst, src = refs1[0]
@@ -377,21 +544,36 @@ def _find_table_reference(t1, t2):
 
 def pql_type(state: State, obj: T.any):
     """Returns the type of the given object
+
+    Example:
+        >> type(1)
+        int
+        >> type([1])
+        list[item: int]
+        >> type(int)
+        type
     """
     return obj.type
 
+
 def pql_repr(state: State, obj: T.any):
-    """Returns the repr of the given object
+    """Returns the representation text of the given object
     """
     try:
-        return new_value_instance(obj.repr(state))
+        return new_value_instance(obj.repr())
     except ValueError:
         value = repr(cast_to_python(state, obj))
         return new_value_instance(value)
 
 def pql_columns(state: State, table: T.collection):
-    """Accepts a table, and returns a dictionary of {column_name: column_type}
+    """Returns a dictionary `{column_name: column_type}` for the given table
+
+    Example:
+        >> columns([0])
+        {item: int}
     """
+    # TODO table: T.table, not collection
+
     elems = table.type.elems
     if isinstance(elems, tuple):    # Create a tuple/list instead of dict?
         elems = {f't{i}':e for i, e in enumerate(elems)}
@@ -399,34 +581,43 @@ def pql_columns(state: State, table: T.collection):
     return ast.Dict_(elems)
 
 
-def pql_cast(state: State, inst: T.any, type: T.type):
-    "Attempt to cast an object to a specified type"
-    type_ = type
+def pql_cast(state: State, obj: T.any, target_type: T.type):
+    """Attempt to cast an object to a specified type
+
+    The resulting object will be of type `target_type`, or a `TypeError`
+    exception will be thrown.
+
+    Parameters:
+        obj: The object to cast
+        target_type: The type to cast to
+
+    """
+    type_ = target_type
     if not isinstance(type_, Type):
-        raise Signal.make(T.TypeError, state, type_, f"Cast expected a type, got {type_} instead.")
+        raise Signal.make(T.TypeError, type_, f"Cast expected a type, got {type_} instead.")
 
-    if inst.type is type_:
-        return inst
+    if obj.type is type_:
+        return obj
 
-    return cast(state, inst, type_)
+    return cast(state, obj, type_)
 
 
-def pql_import_table(state: State, name: T.string, columns: T.list[T.string] = objects.null):
-    """Import an existing table from SQL
+def pql_import_table(state: State, name: T.string, columns: T.list[T.string].as_nullable() = objects.null):
+    """Import an existing table from the database, and fill in the types automatically.
 
-    If the columns argument is provided, only these columns will be imported.
+    Parameters:
+        name: The name of the table to import
+        columns: If this argument is provided, only these columns will be imported.
 
     Example:
         >> import_table("my_sql_table", ["some_column", "another_column])
     """
     name_str = cast_to_python(state, name)
-    columns_whitelist = cast_to_python(state, columns) or []
-    if not isinstance(columns_whitelist, list):
-        raise Signal.make(T.TypeError, state, columns, "Expected list")
-    if not isinstance(name_str, str):
-        raise Signal.make(T.TypeError, state, name, "Expected string")
-
-    columns_whitelist = set(columns_whitelist)
+    assert isinstance(name_str, str)
+    columns_whitelist = cast_to_python(state, columns)
+    if columns_whitelist is not None:
+        assert isinstance(columns_whitelist, list)
+        columns_whitelist = set(columns_whitelist)
 
     # Get table type
     t = state.db.import_table_type(state, name_str, columns_whitelist)
@@ -440,8 +631,13 @@ def pql_import_table(state: State, name: T.string, columns: T.list[T.string] = o
 def pql_connect(state: State, uri: T.string, load_all_tables: T.bool = ast.false, auto_create: T.bool = ast.false):
     """Connect to a new database, specified by the uri
 
-    If load_all_tables is true, loads all the tables in the database to the global namespace.
-    If auto_create is true, create the database if it doesn't exist (Sqlite only)
+    Parameters:
+        uri: A string specifying which database to connect to (e.g. "sqlite:///test.db")
+        load_all_tables: If true, loads all the tables in the database into the global namespace.
+        auto_create: If true, creates the database if it doesn't already exist (Sqlite only)
+
+    Example:
+        >> connect("sqlite://:memory:")     // Connect to a database in memory
     """
     uri = cast_to_python(state, uri)
     load_all_tables = cast_to_python(state, load_all_tables)
@@ -452,7 +648,7 @@ def pql_connect(state: State, uri: T.string, load_all_tables: T.bool = ast.false
     return objects.null
 
 def pql_help(state: State, inst: T.any = objects.null):
-    """Provides a brief summary for a given object
+    """Provides a brief summary for the given object
     """
     if inst is objects.null:
         text = (
@@ -468,16 +664,18 @@ def pql_help(state: State, inst: T.any = objects.null):
 
 
     lines = []
-    if isinstance(inst, objects.Function):
-        lines = ['', inst.help_str(state),'']
-        doc = inst.docstring
+    try:
+        doc = autodoc(inst).print_text()    # TODO maybe html
         if doc:
             lines += [doc]
-    else:
-        raise Signal.make(T.TypeError, state, None, "help() only accepts functions at the moment")
+    except NotImplementedError:
+        lines = [f"No help available for {inst.repr()}"]
+    except runtype.DispatchError:
+        lines += [f"<doc not available yet for object of type '{inst.type}>'"]
+
 
     text = '\n'.join(lines) + '\n'
-    return new_value_instance(text).replace(type=T.text)
+    return new_value_instance(text).replace(type=T._rich)
 
 def _get_doc(v):
     s = ''
@@ -505,7 +703,10 @@ def pql_names(state: State, obj: T.any = objects.null):
 
 
 def pql_tables(state: State):
-    "Return a table of all tables in the database"
+    """Returns a table of all the persistent tables in the database.
+
+    The resulting table has two columns: name, and type.
+    """
     names = state.db.list_tables()
     values = [(name, state.db.import_table_type(state, name, None)) for name in names]
     tuples = [sql.Tuple(T.list[T.string], [new_str(n).code,new_str(t).code]) for n,t in values]
@@ -514,6 +715,17 @@ def pql_tables(state: State):
     return objects.new_const_table(state, table_type, tuples)
 
 
+def pql_env_vars(state: State):
+    """Returns a table of all the environment variables.
+
+    The resulting table has two columns: name, and value.
+    """
+    tuples = [sql.Tuple(T.list[T.string], [new_str(n).code,new_str(t).code]) for n,t in os.environ.items()]
+
+    table_type = T.table({'name':T.string, 'value': T.string})
+    return objects.new_const_table(state, table_type, tuples)
+
+
 
 def create_internal_funcs(d):
     new_d = {}
@@ -529,21 +741,77 @@ breakpoint_funcs = create_internal_funcs({
 })
 
 
-def pql_exit(state, value: T.int.as_nullable() = None):
+def pql_exit(state, value: T.any.as_nullable() = None):
     """Exit the current interpreter instance.
 
     Can be used from running code, or the REPL.
 
     If the current interpreter is nested within another Preql interpreter (e.g. by using debug()),
     exit() will return to the parent interpreter.
+
     """
     raise ExitInterp(value)
 
 
 
+def import_pandas(state, dfs):
+    """Import pandas.DataFrame instances into SQL tables
+    """
+    import pandas as pd
+    def normalize_item(i):
+        if pd.isna(i):
+            return None
+        i = i.item() if hasattr(i, 'item') else i
+        return i
+
+    for name, df in dfs.items():
+        if isinstance(df, pd.Series):
+            cols = ['key', 'value']
+            rows = [(dt.to_pydatetime() if isinstance(dt, datetime) else dt,v) for dt, v in df.items()]
+        else:
+            assert isinstance(df, pd.DataFrame)
+            cols = list(df)
+            rows = [[normalize_item(i) for i in rec]
+                    for rec in df.to_records()]
+            rows = [ row[1:] for row in rows ]    # drop index
+
+        yield new_table_from_rows(state, name, cols, rows)
+
+def pql_import_json(state: State, table_name: T.string, uri: T.string):
+    """Imports a json file into a new table.
+
+    Returns the newly created table.
+
+    Parameters:
+        table_name: The name of the table to create
+        uri: A path or URI to the JSON file
+
+    Note:
+        This function requires the `pandas` Python package.
+    """
+    table_name = cast_to_python(state, table_name)
+    uri = cast_to_python(state, uri)
+    print(f"Importing JSON file: '{uri}'")
+
+    import pandas
+    df = pandas.read_json(uri)
+    tbl ,= import_pandas(state, {table_name: df})
+    return tbl
+
+
+
 def pql_import_csv(state: State, table: T.table, filename: T.string, header: T.bool = ast.Const(T.bool, False)):
-    "Import a csv into an existing table"
+    """Import a csv file into an existing table
+
+    Parameters:
+        table: A table into which to add the rows.
+        filename: A path to the csv file
+        header: If true, skips the first line
+    """
     # TODO better error handling, validation
+    from tqdm import tqdm   # TODO use rich instead?
+
+
     filename = cast_to_python(state, filename)
     header = cast_to_python(state, header)
     print(f"Importing CSV file: '{filename}'")
@@ -561,8 +829,8 @@ def pql_import_csv(state: State, table: T.table, filename: T.string, header: T.b
 
     with open(filename, 'r', encoding='utf8') as f:
         line_count = len(list(f))
+        f.seek(0)
 
-    with open(filename, 'r', encoding='utf8') as f:
         reader = csv.reader(f)
         for i, row in enumerate(tqdm(reader, total=line_count)):
             if i == 0:
@@ -612,19 +880,35 @@ def _rest_table_endpoint(state, table):
 
 
 def pql_serve_rest(state: State, endpoints: T.struct, port: T.int = new_int(8080)):
-    "Start a starlette server that exposes the current namespace as REST API"
+    """Start a starlette server (HTTP) that exposes the current namespace as REST API
+
+    Parameters:
+        endpoints: A struct of type (string => function), mapping names to the functions.
+        port: A port from which to serve the API
+
+    Note:
+        Requires the `starlette` package for Python. Run `pip install starlette`.
+
+    Example:
+        >> func index() = "Hello World!"
+        >> serve_rest({index: index})
+        INFO     Started server process [85728]
+        INFO     Waiting for application startup.
+        INFO     Application startup complete.
+        INFO     Uvicorn running on http://127.0.0.1:8080 (Press CTRL+C to quit)
+    """
 
     try:
         from starlette.applications import Starlette
         from starlette.responses import JSONResponse
         from starlette.routing import Route
     except ImportError:
-        raise Signal.make(T.ImportError, state, None, "starlette not installed! Run 'pip install starlette'")
+        raise Signal.make(T.ImportError, None, "starlette not installed! Run 'pip install starlette'")
 
     try:
         import uvicorn
     except ImportError:
-        raise Signal.make(T.ImportError, state, None, "uvicorn not installed! Run 'pip install uvicorn'")
+        raise Signal.make(T.ImportError, None, "uvicorn not installed! Run 'pip install uvicorn'")
 
     port_ = cast_to_python(state, port)
 
@@ -645,7 +929,7 @@ def pql_serve_rest(state: State, endpoints: T.struct, port: T.int = new_int(8080
         elif func.type <= T.collection:
             routes.append(Route(path, endpoint=_rest_table_endpoint(state, func)))
         else:
-            raise Signal.make(T.TypeError, state, func, f"Expected a function or a table, got {func.type}")
+            raise Signal.make(T.TypeError, func, f"Expected a function or a table, got {func.type}")
 
     app = Starlette(debug=True, routes=routes)
 
@@ -658,6 +942,7 @@ internal_funcs = create_internal_funcs({
     'help': pql_help,
     'names': pql_names,
     'tables': pql_tables,
+    'env_vars': pql_env_vars,
     'dir': pql_names,
     'connect': pql_connect,
     'import_table': pql_import_table,
@@ -680,6 +965,7 @@ internal_funcs = create_internal_funcs({
     'cast': pql_cast,
     'columns': pql_columns,
     'import_csv': pql_import_csv,
+    'import_json': pql_import_json,
     'serve_rest': pql_serve_rest,
     'force_eval': pql_force_eval,
     'fmt': pql_fmt,
