@@ -12,7 +12,7 @@ from . import sql
 from . import pql_types
 
 from .pql_types import ITEM_NAME, T, Type, Object
-from .types_impl import flatten_type, join_names, pql_repr
+from .types_impl import flatten_type, join_names, pql_repr, kernel_type
 
 
 # Functions
@@ -62,6 +62,11 @@ class Module(Object):
     def __repr__(self):
         return f'<preql:Module | {len(self.namespace)} members>'
 
+    def public_functions(self):
+        funcs = [v for v in self.namespace.values() if v.type <= T.function and v.docstring and not v.name.startswith('_')]
+        funcs.sort(key=lambda f: f.name)
+        return funcs
+
 
 class Function(Object):
 
@@ -69,34 +74,46 @@ class Function(Object):
     def type(self):
         return T.function[tuple(p.type or T.any for p in self.params)](param_collector=self.param_collector is not None)
 
-    def help_str(self, state):
-        raise NotImplementedError()
+    def help_str(self):
+        params = []
+        for p in self.params:
+            s = p.name
+            if p.type:
+                s += f": {p.type}"
+            if p.default:
+                s += f"={p.default.repr()}"
+            params.append(s)
+
+        if self.param_collector is not None:
+            params.append(f"...{self.param_collector.name}")
+        param_str = ', '.join(params)
+        return f"func {self.name}({param_str}) = ..."
 
     def __repr__(self):
         return f'<preql:Function | {self.name}: {self.type}>'
 
     @listgen
-    def match_params_fast(self, state, args):
+    def match_params_fast(self, args):
         for i, p in enumerate(self.params):
             if i < len(args):
                 v = args[i]
             else:
                 v = p.default
                 if v is None:
-                    raise Signal.make(T.TypeError, state, None, f"Function '{self.name}' is missing a value for parameter '{p.name}'")
+                    raise Signal.make(T.TypeError, None, f"Function '{self.name}' is missing a value for parameter '{p.name}'")
 
 
             yield p, v
 
 
-    def _localize_keys(self, state, struct):
+    def _localize_keys(self, _state, struct):
         raise NotImplementedError()
 
     def match_params(self, state, args):
 
         # If no keyword arguments, matching is much simpler and faster
         if all(not isinstance(a, (ast.NamedField, ast.Ellipsis)) for a in args):
-            return self.match_params_fast(state, args)
+            return self.match_params_fast(args)
 
         # Canonize args for the rest of the function
         inline_args = []
@@ -109,9 +126,10 @@ class Function(Object):
                     raise NotImplementedError("Cannot exclude keys when inlining struct")
 
                 # XXX we only want to localize the keys, not the values
+                # TODO remove this?
                 d = self._localize_keys(state, a.from_struct)
                 if not isinstance(d, dict):
-                    raise Signal.make(T.TypeError, state, None, f"Expression to inline is not a map: {d}")
+                    raise Signal.make(T.TypeError, None, f"Expression to inline is not a map: {d}")
                 for k, v in d.items():
                     inline_args.append(ast.NamedField(k, new_value_instance(v)))
             else:
@@ -126,11 +144,11 @@ class Function(Object):
         else:
             if not all(n for n in named[first_named:]):
                 # TODO meta
-                raise Signal.make(T.TypeError, state, None, f"Function {self.name} received a non-named argument after a named one!")
+                raise Signal.make(T.TypeError, None, f"Function {self.name} received a non-named argument after a named one!")
 
         if first_named > len(self.params):
             # TODO meta
-            raise Signal.make(T.TypeError, state, None, f"Function '{self.name}' takes {len(self.params)} parameters but received {first_named} arguments.")
+            raise Signal.make(T.TypeError, None, f"Function '{self.name}' takes {len(self.params)} parameters but received {first_named} arguments.")
 
         values = {p.name: p.default for p in self.params}
 
@@ -149,13 +167,13 @@ class Function(Object):
                     collected[arg_name] = named_arg.value
                 else:
                     # TODO meta
-                    raise Signal.make(T.TypeError, state, None, f"Function '{self.name}' has no parameter named '{arg_name}'")
+                    raise Signal.make(T.TypeError, None, f"Function '{self.name}' has no parameter named '{arg_name}'")
 
 
         for name, value in values.items():
             if value is None:
                 # TODO meta
-                raise Signal.make(T.TypeError, state, None, f"Error calling function '{self.name}': parameter '{name}' has no value")
+                raise Signal.make(T.TypeError, None, f"Error calling function '{self.name}': parameter '{name}' has no value")
 
         matched = [(p, values.pop(p.name)) for p in self.params]
         assert not values, values
@@ -236,7 +254,7 @@ class Instance(AbsInstance):
     def make(cls, code, type_, instances, *extra):
         return cls(code, type_, merge_subqueries(instances), *extra)
 
-    def repr(self, state):
+    def repr(self):
         # Overwritten in evaluate.py
         raise NotImplementedError()
         #     return f'<instance of {self.type.repr(state)}>'
@@ -274,8 +292,8 @@ def new_value_instance(value, type_=None, force_type=False):
 class ValueInstance(Instance):
     local_value: object
 
-    def repr(self, state):
-        return pql_repr(state, self.type, self.local_value)
+    def repr(self):
+        return pql_repr(self.type, self.local_value)
 
     @property
     def value(self):
@@ -292,7 +310,7 @@ class TableInstance(CollectionInstance):
 
     @property
     def __columns(self):
-        return {n:self.get_column(n) for n in self.type.elems}
+        return {n:self.get_column(n) for n in self.type.elems.keys()}
 
     def get_column(self, name):
         # TODO memoize? columns shouldn't change
@@ -300,9 +318,8 @@ class TableInstance(CollectionInstance):
         return make_instance_from_name(t[name], name) #t.column_codename(name))
 
     def all_attrs(self):
-        # XXX hacky way to write it
-        attrs = dict(self.type.methods)
-        return SafeDict(attrs).update(self.__columns)
+        attrs = SafeDict(self.type.methods)
+        return attrs.update(self.__columns)
 
     def get_attr(self, name):
         try:
@@ -314,33 +331,6 @@ class TableInstance(CollectionInstance):
             except KeyError:
                 raise pql_AttributeError(name)
 
-@dataclass
-class ListInstance(CollectionInstance):
-    def __post_init__(self):
-        assert self.type <= T.list, self.type
-
-    def get_column(self, name):
-        # TODO memoize? columns shouldn't change
-        assert name == ITEM_NAME
-        t = self.type
-        return make_instance_from_name(t.elem, name)
-
-    def all_attrs(self):
-        # XXX hacky way to write it
-        attrs = dict(self.type.methods)
-        attrs[ITEM_NAME] = self.get_column(ITEM_NAME)
-        return attrs
-
-    def get_attr(self, name):
-        if name == ITEM_NAME:
-            v = self.type.elem
-            return SelectedColumnInstance(self, v, name)
-        else:
-            try:
-                return MethodInstance(self, self.type.methods[name])
-            except KeyError:
-                raise pql_AttributeError(name)
-
 
 
 def make_instance_from_name(t, cn):
@@ -350,9 +340,7 @@ def make_instance_from_name(t, cn):
 
 def make_instance(code, t, insts):
     assert not t.issubtype(T.struct), t
-    if t <= T.list:
-        return ListInstance.make(code, t, insts)
-    elif t <= T.table:
+    if t <= T.table:
         return TableInstance.make(code, t, insts)
     elif t <= T.aggregate:
         return AggregateInstance(t, make_instance(code, t.elem, insts))
@@ -394,15 +382,14 @@ class AbsStructInstance(AbsInstance):
     def get_attr(self, name):
         if name in self.attrs:
             attr = self.attrs[name]
-            if self.type <= T.vectorized:
-                attr = vectorized(attr)
-            return attr
+            return inherit_vectorized(attr, [self])
         else:
             raise pql_AttributeError(name)
 
     @property
     def code(self):
         # XXX this shouldn't even be allowed to happen in the first place
+        breakpoint()
         raise Signal(T.TypeError, [], "structs are abstract objects and cannot be sent to target. Choose one of its members instead.")
 
 
@@ -412,7 +399,7 @@ class StructInstance(AbsStructInstance):
     attrs: Dict[str, Object]
 
     def __post_init__(self):
-        assert self.type <= T.union[T.struct, T.vectorized[T.struct]]
+        assert kernel_type( self.type ) <= T.struct
 
     @property
     def subqueries(self):
@@ -428,8 +415,8 @@ class StructInstance(AbsStructInstance):
     def all_attrs(self):
         return self.attrs
 
-    def repr(self, state):
-        attrs = [f'{k}: {v.repr(state)}' for k, v in self.attrs.items()]
+    def repr(self):
+        attrs = [f'{k}: {v.repr()}' for k, v in self.attrs.items()]
         return '{%s}' % ', '.join(attrs)
 
 
@@ -461,8 +448,8 @@ class MapInstance(AbsStructInstance):
     def primary_key(self):
         return self
 
-    def repr(self, state):
-        inner = [f'{name}: {v.repr(state)}' for name, v in self.attrs.items()]
+    def repr(self):
+        inner = [f'{name}: {v.repr()}' for name, v in self.attrs.items()]
         return 'Map{%s}' % ', '.join(inner)
 
 
@@ -474,8 +461,8 @@ class RowInstance(StructInstance):
             # XXX this is a hack!
             return list(self.attrs.values())[0]
 
-    def repr(self, state):
-        inner = [f'{name}: {v.repr(state)}' for name, v in self.attrs.items()]
+    def repr(self):
+        inner = [f'{name}: {v.repr()}' for name, v in self.attrs.items()]
         return 'Row{%s}' % ', '.join(inner)
 
 
@@ -524,8 +511,11 @@ class SelectedColumnInstance(AbsInstance):
     def _resolve_attr(self):
         return self.parent.get_column(self.name)
 
-    def repr(self, state):
-        p = self.parent.repr(state)
+    def repr(self):
+        try:
+            p = self.parent.type.options['name'].repr_name
+        except KeyError:
+            p = '?'
         return f'{p}.{self.name}'
 
 
@@ -548,32 +538,28 @@ def vectorized(inst):
     return inst.replace(type=T.vectorized[inst.type])
 
 def unvectorized(inst):
-    assert inst.type <= T.vectorized
-    return inst.replace(type=inst.type.elem)
+    if inst.type <= T.vectorized:
+        return inst.replace(type=inst.type.elem)
+    return inst
 
+def inherit_vectorized_type(t, objs):
+    for src in objs:
+        if src.type <= T.vectorized:
+            return T.vectorized[t]
+    return t
 
-@dy
-def unvectorize_args(x: list):
-    if not x:
-        return False, x
-    was_vec, objs = zip(*[unvectorize_args(i) for i in x])
-    return any(was_vec), list(objs)
+def inherit_vectorized(o, objs):
+    for src in objs:
+        if src.type <= T.vectorized:
+            return vectorized(o)
+    return o
 
-@dy
-def unvectorize_args(x: AbsInstance):
-    if x.type <= T.vectorized:
-        return True, unvectorized(x)
-    return False, x
-
-@dy
-def unvectorize_args(x):
-    return False, x
 
 
 null = ValueInstance.make(sql.null, T.nulltype, [], None)
 
 @dataclass
-class EmptyListInstance(ListInstance):
+class EmptyListInstance(TableInstance):
     """Special case, because it is untyped
     """
 
@@ -597,12 +583,8 @@ def alias_table_columns(t, prefix):
 
 def new_table(type_, name=None, instances=None, select_fields=False):
     "Create new table instance"
-    if type_ <= T.list:
-        cls = ListInstance
-    else:
-        cls = TableInstance
     name = name if name else type_.options.get('name', sql.Id('anon'))
-    inst = cls.make(sql.TableName(type_, name), type_, instances or [])
+    inst = TableInstance.make(sql.TableName(type_, name), type_, instances or [])
 
     if select_fields:
         code = sql.Select(type_, inst.code, [sql.Name(t, n) for n, t in type_.elems.items()])
@@ -642,6 +624,8 @@ def from_python(value):
         return ast.Const(T.bool, value)
     elif isinstance(value, int):
         return ast.Const(T.int, value)
+    elif isinstance(value, float):
+        return ast.Const(T.float, value)
     elif isinstance(value, list):
         # return ast.List_(T.list[T.any], list(map(from_python, value)))
         return PythonList(value)
