@@ -8,6 +8,8 @@ loop = asyncio.SelectorEventLoop(selector)
 asyncio.set_event_loop(loop)
 ### XXX End of fix
 
+from pathlib import Path
+
 from pygments.lexers.go import GoLexer
 from prompt_toolkit import PromptSession
 from prompt_toolkit.lexers import PygmentsLexer
@@ -28,6 +30,7 @@ from .parser import parse_stmts
 from .loggers import repl_log
 from . import settings
 from .autocomplete import autocomplete
+from .context import context
 
 
 
@@ -158,6 +161,9 @@ class PreqlStyle(Style):
         Error:                  'bg:ansired ansigray',
     }
 
+from prompt_toolkit.history import FileHistory
+from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
+
 def start_repl(p, prompt=' >> '):
     save_last = '_'   # XXX A little hacky
 
@@ -173,6 +179,8 @@ def start_repl(p, prompt=' >> '):
             completer=Autocompleter(p.interp.state),
             # key_bindings=kb
             validator=MyValidator(),
+            history=FileHistory(str(Path.home() / '.preql_history')),
+            auto_suggest=AutoSuggestFromHistory(),
         )
 
         @Condition
@@ -191,7 +199,8 @@ def start_repl(p, prompt=' >> '):
                 start_time = time()
                 try:
                     if code == '.':
-                        console.print(table_more(p.interp.state), overflow='ellipsis')
+                        with context(state=p.interp.state):
+                            console.print(table_more(), overflow='ellipsis')
                         continue
 
                     # Evaluate (Really just compile)
@@ -204,7 +213,9 @@ def start_repl(p, prompt=' >> '):
                         if save_last:
                             p.interp.set_var(save_last, res)
 
-                        res_repr = res.repr(p.interp.state)
+                        with context(state=p.interp.state):
+                            res_repr = res.repr()
+
                         # repl_log.info(res)
                         if isinstance(res_repr, str) and res.type == T.string:  # Not text
                             if len(res_repr) > 200:
@@ -217,7 +228,7 @@ def start_repl(p, prompt=' >> '):
                     # p.interp.set_var('_e', objects.ExceptionInstance(e))
                     continue
                 except ExitInterp as e:
-                    return
+                    return e.value
                 except Exception as e:
                     repl_log.exception(e)
                     raise
