@@ -3,10 +3,10 @@ import json
 from datetime import datetime
 from typing import List, Optional, Dict
 
-from .utils import dataclass, X, listgen, field_list
+from .utils import dataclass, X, listgen, field_list, safezip
 from . import pql_types
-from .pql_types import ITEM_NAME, T, Type, dp_type, dp_inst, Id
-from .types_impl import join_names, flatten_type
+from .pql_types import T, Type, dp_type, dp_inst, Id
+from .types_impl import join_names, flatten_type, kernel_type
 from .exceptions import Signal
 
 duck = 'duck'
@@ -72,7 +72,10 @@ class SqlTree(Sql):
     def finalize_with_subqueries(self, qb, subqueries):
         if subqueries:
             subqs = [q.compile_wrap(qb).finalize(qb) for (name, q) in subqueries.items()]
-            sql_code = ['WITH RECURSIVE ']
+            if qb.target in (postgres, mysql):
+                sql_code = ['WITH RECURSIVE ']
+            else:
+                sql_code = ['WITH ']
             sql_code += join_comma([q, '\n    '] for q in subqs)
         else:
             sql_code = []
@@ -107,7 +110,7 @@ class CompiledSQL(Sql):
         else:
             if self._is_select and not self._needs_select:
                 # Bad recursion
-                if qb.target == 'postgres' or qb.target == 'mysql' or qb.target == 'duck':
+                if qb.target in (postgres, mysql, duck):
                     code = ['('] + code + [') ', qb.unique_name()]  # postgres requires an alias
                 else:
                     code = ['('] + code + [')']
@@ -204,7 +207,7 @@ class EmptyList(Table):
     _is_select = True
 
     def _compile(self, qb):
-        return ['SELECT NULL AS VALUE LIMIT 0']
+        return ['SELECT NULL AS ITEM LIMIT 0']
 
 @dataclass
 class TableName(Table):
@@ -239,6 +242,20 @@ class CountTable(Scalar):
     def _compile(self, qb):
         return [f'(SELECT COUNT(*) FROM '] + self.table.compile_wrap(qb).code + [')']
 
+@dataclass
+class JsonLength(Scalar):
+    expr: Sql
+    type = T.int
+
+    def _compile(self, qb):
+        code = self.expr.compile_wrap(qb).code
+        if qb.target == sqlite:
+            return [f'(length('] + code + [') - length(replace('] + code + [f', "{_ARRAY_SEP}", ""))) / length("{_ARRAY_SEP}") + 1']
+        elif qb.target == postgres:
+            return [f'array_length('] + code + [', 1)']
+        else:
+            return [f'json_length('] + code + [')']
+
 @dataclass
 class FuncCall(SqlTree):
     type: Type
@@ -252,11 +269,11 @@ class FuncCall(SqlTree):
 @dataclass
 class Cast(SqlTree):
     type: Type
-    as_type: str
     value: Sql
 
     def _compile(self, qb):
-        return [f'CAST('] + self.value.compile_wrap(qb).code + [f' AS {self.as_type})']
+        t = compile_type(qb.target, self.type.as_nullable())     # XXX as-nullable here is a hack
+        return [f'CAST('] + self.value.compile_wrap(qb).code + [f' AS {t})']
 
 
 @dataclass
@@ -291,6 +308,8 @@ class MakeArray(SqlTree):
             return ['array_agg('] + field + [')']
         elif qb.target == mysql:
             return ['json_arrayagg('] + field + [')']
+        elif qb.target == bigquery:
+            return ['array_agg('] + field + [')']
 
         assert False, qb.target
 
@@ -499,14 +518,16 @@ class InsertConsts(SqlTree):
     type = T.nulltype
 
     def _compile(self, qb):
+        cols = self.cols
+        tuples = self.tuples
         assert self.tuples, self
 
         values = join_comma(
             parens(join_comma([e.compile_wrap(qb).code for e in tpl]))
-            for tpl in self.tuples
+            for tpl in tuples
         )
 
-        cols = [qb.quote(Id(c)) for c in self.cols]
+        cols = [qb.quote(Id(c)) for c in cols]
         q = ['INSERT INTO', qb.safe_name(self.table),
              "(", ', '.join(cols), ")",
              "VALUES ",
@@ -558,13 +579,22 @@ class SelectValue(Atom, TableOperation):
 
     type = property(X.value.type)
 
+# @dataclass
+# class RowDict(SqlTree):
+#     values: Dict[str, Sql]
+
+#     def _compile(self, qb):
+#         breakpoint()
+#         return {v.compile_wrap(qb).code + [f' as {name}'] for name, v in self.values.items()}
+
+
 @dataclass
-class RowDict(SqlTree):
-    values: Dict[str, Sql]
-
+class ValuesTuple(SqlTree):
+    type: Type
+    values: List[Sql]
     def _compile(self, qb):
-        return {v.compile_wrap(qb).code + [f' as {name}'] for name, v in self.values.items()}
-
+        values = [v.compile_wrap(qb) for v in self.values]
+        return join_comma(v.code for v in values)
 
 @dataclass
 class Values(Table):
@@ -720,11 +750,12 @@ class Select(TableOperation):
 @listgen
 def join_sep(code_list, sep):
     code_list = list(code_list)
-    yield from code_list[0]
-    for c in code_list[1:]:
-        assert isinstance(c, list)
-        yield sep
-        yield from c
+    if code_list:
+        yield from code_list[0]
+        for c in code_list[1:]:
+            assert isinstance(c, list)
+            yield sep
+            yield from c
 
 def join_comma(code_list):
     return join_sep(code_list, ", ")
@@ -732,14 +763,17 @@ def join_comma(code_list):
 @dataclass
 class Subquery(SqlTree):
     table_name: str
-    fields: List[Name]
+    fields: Optional[List[Name]]
     query: Sql
     type = property(X.query.type)
 
     def _compile(self, qb):
         query = self.query.compile(qb).code
-        fields = [f.compile_wrap(qb.replace(is_root=False)).code for f in self.fields]
-        fields_str = ["("] + join_comma(fields) + [")"] if fields else []
+        if qb.target == bigquery:
+            fields_str = []
+        else:
+            fields = [f.compile_wrap(qb.replace(is_root=False)).code for f in self.fields]
+            fields_str = ["("] + join_comma(fields) + [")"] if fields else []
         return [f"{self.table_name}"] + fields_str + [" AS ("] + query + [")"]
 
 def _enum_is_last(seq):
@@ -797,16 +831,54 @@ def _sql_primitive_to_float(e):
 
 def create_list(name, elems):
     # Assumes all elems have the same type!
-    list_type = T.list[elems[0].type]
-    fields = [Name(list_type.elem, ITEM_NAME)]
-    subq = Subquery(name, fields, Values(list_type, elems))
-    table = TableName(list_type, Id(name))
-    return table, subq, list_type
+    t = T.list[elems[0].type]
+    subq = TableQueryValues(t, name, elems)
+    table = TableName(t, Id(name))
+    return table, subq, t
 
-def create_table(table_type, name, rows):
-    fields = [Name(col_type, col_name) for col_name, col_type in table_type.elems.items()]
-    subq = Subquery(name, fields, Values(table_type, rows))
-    table = TableName(table_type, Id(name))
+
+@dataclass
+class BigQueryValues(SqlTree):
+    type: Type
+    values: List[Sql]
+
+    def _compile(self, qb):
+        cols = list(self.type.elems)
+
+        rows = [
+            (
+             ['STRUCT('] + join_comma(v.compile(qb).code + [" as ", name]
+                                  for name, v in safezip(cols, row.values)) + [")"]
+            ) if isinstance(row, Tuple) else row.compile(qb).code
+             for row in self.values
+        ]
+
+        return ["SELECT * FROM UNNEST(["] + join_comma(rows) + ["]) as item"]
+
+
+
+@dataclass
+class TableQueryValues(SqlTree):
+    "A subquery which returns a table of given values"
+
+    type: Type
+    name: str
+    rows: list
+
+    def _compile(self, qb):
+        if qb.target != 'bigquery':
+            values_cls = Values
+            fields = [Name(col_type, col_name) for col_name, col_type in self.type.elems.items()]
+        else:
+            values_cls = BigQueryValues
+            fields = None
+
+        subq = Subquery(self.name, fields, values_cls(self.type, self.rows))
+        return subq._compile(qb)
+
+def create_table(t, name, rows):
+    subq = TableQueryValues(t, name, rows)
+    table = TableName(t, Id(name))
     return table, subq
 
 def table_slice(table, start, stop):
@@ -829,12 +901,18 @@ def arith(target, res_type, op, args):
     elif op == '/':
         if target != mysql:
             # In MySQL division returns a float. All others return int
-            arg_codes[0] = Cast(T.float, 'float', arg_codes[0])
+            arg_codes[0] = Cast(T.float, arg_codes[0])
     elif op == '/~':
         if target == mysql:
             op = 'DIV'
+        elif target == bigquery:
+            # safe div?
+            return FuncCall(res_type, 'div', arg_codes)
         else:
             op = '/'
+    elif op == '%':
+        if target is bigquery:
+            return FuncCall(res_type, 'mod', arg_codes)
     elif op == '**':
         return FuncCall(T.float, 'power', arg_codes)
 
@@ -863,7 +941,7 @@ class StringSlice(SqlTree):
             params = string + [', '] + start
             if length:
                 params += [', '] + length
-        elif qb.target in (postgres, mysql):
+        elif qb.target in (postgres, mysql, bigquery):
             f = 'substring'
             params = string + [' from '] + start
             if length:
@@ -887,9 +965,13 @@ def _repr(t: T.datetime, x):
     # TODO Better to pass the object instead of a string?
     return repr(str(x))
 
+from .context import context
 @dp_type
 def _repr(t: T.union[T.string, T.text], x):
-    return "'%s'" % str(x).replace("'", "''")
+    if context.state.db.target == bigquery:
+        return "'%s'" % str(x).replace("'", r"\'")
+    else:
+        return "'%s'" % str(x).replace("'", "''")
 
 def make_value(x):
     if x is None:
@@ -929,25 +1011,28 @@ def compile_type_def(state, table_name, table) -> Sql:
             if target == postgres:
                 type_ = "SERIAL" # Postgres
             elif target == mysql:
-                type_ = "INT NOT NULL AUTO_INCREMENT"
+                type_ = "INTEGER NOT NULL AUTO_INCREMENT"
+            elif target == bigquery:
+                type_ = "STRING NOT NULL"
             else:
                 type_ = "INTEGER"   # TODO non-int idtypes
         else:
-            type_ = compile_type(c)
+            type_ = compile_type(target if target != mysql else 'mysql_def', c)
 
         columns.append( f'{_quote(target, name)} {type_}' )
 
         if c <= T.t_relation:
-            # TODO any column, using projection / get_attr
-            if not table.options.get('temporary', False):
-                # In postgres, constraints on temporary tables may reference only temporary tables
-                rel = c.options['rel']
-                if rel['key']:          # Requires a unique constraint
-                    _tbl_name ,= rel['table'].options['name'].parts   # TODO fix for multiple parts
-                    s = f"FOREIGN KEY({name}) REFERENCES {_quote(target, _tbl_name)}({rel['column']})"
-                    posts.append(s)
+            if state.db.target != 'bigquery':
+                # TODO any column, using projection / get_attr
+                if not table.options.get('temporary', False):
+                    # In postgres, constraints on temporary tables may reference only temporary tables
+                    rel = c.options['rel']
+                    if rel['key']:          # Requires a unique constraint
+                        _tbl_name ,= rel['table'].options['name'].parts   # TODO fix for multiple parts
+                        s = f"FOREIGN KEY({name}) REFERENCES {_quote(target, _tbl_name)}({rel['column']})"
+                        posts.append(s)
 
-    if pks:
+    if pks and target != bigquery:
         names = ", ".join(pks)
         posts.append(f"PRIMARY KEY ({names})")
 
@@ -960,38 +1045,64 @@ def compile_drop_table(state, table_name) -> Sql:
     return RawSql(T.nulltype, f'DROP TABLE {_quote(target, table_name)}')
 
 @dp_type
-def compile_type(type_: T.t_relation):
+def compile_type(target, type_: T.t_relation):
     # TODO might have a different type
     #return 'INTEGER'    # Foreign-key is integer
-    return compile_type(type_.elems['item'])
+    return compile_type(target, type_.elems['item'])
 
 @dp_type
-def compile_type(type_: T.primitive):
-    assert type_ <= T.primitive
-    s = {
-        'int': "INTEGER",
-        'string': "VARCHAR(4000)",
-        'float': "FLOAT",
-        'bool': "BOOLEAN",
-        'text': "TEXT",
-        # 't_relation': "INTEGER",
-        'datetime': "TIMESTAMP",
-    }[type_.typename]
+def compile_type(target, type_: T.primitive):
+    if target == bigquery:
+        d = {
+            'int': "INT64",
+            'string': "STRING",
+            'float': "FLOAT64",
+            'bool': "BOOLEAN",
+            'text': "STRING",
+            'datetime': "TIMESTAMP",
+        }
+    elif target == mysql:
+        d = {
+            'int': "SIGNED",
+            'string': "VARCHAR(4000)",
+            'float': "FLOAT",
+            'bool': "BOOLEAN",
+            'text': "TEXT",
+            'datetime': "TIMESTAMP",
+        }
+    else:
+        d = {
+            'int': "INTEGER",
+            'string': "VARCHAR(4000)",
+            'float': "FLOAT",
+            'bool': "BOOLEAN",
+            'text': "TEXT",
+            'datetime': "TIMESTAMP",
+        }
+    s = d[type_.typename]
     if not type_.maybe_null():
         s += " NOT NULL"
     return s
 
 @dp_type
-def compile_type(_type: T.nulltype):
+def compile_type(target, _type: T.nulltype):
+    if target == bigquery:
+        return 'INT64'
     return 'INTEGER'    # TODO is there a better value here? Maybe make it read-only somehow
 
 @dp_type
-def compile_type(idtype: T.t_id):
-    s = "INTEGER"   # TODO non-int idtypes
+def compile_type(target, idtype: T.t_id):
+    if target == bigquery:
+        s = "STRING"
+    else:
+        s = "INTEGER"   # TODO non-int idtypes
     if not idtype.maybe_null():
         s += " NOT NULL"
     return s
 
+@dp_type
+def compile_type(target, _type: T.json):
+    return 'JSON'
 
 
 
@@ -1041,9 +1152,14 @@ def _from_sql_primitive(p):
 
 @dp_inst
 def from_sql(state, arr: T.list):
-    if not all(len(e)==1 for e in arr.value):
+    fields = flatten_type(arr.type)
+    if not all(len(e)==len(fields) for e in arr.value):
         raise Signal.make(T.TypeError, state, None, f"Expected 1 column. Got {len(arr.value[0])}")
-    return [_from_sql_primitive(e[0]) for e in arr.value]
+
+    if arr.type.elem <= T.struct:
+        return [{n: _from_sql_primitive(e) for (n, _t), e in safezip(fields, tpl)} for tpl in arr.value]
+    else:
+        return [_from_sql_primitive(e[0]) for e in arr.value]
 
 @dp_inst
 @listgen
@@ -1053,7 +1169,11 @@ def from_sql(state, arr: T.table):
         if len(row) != expected_length:
             raise Signal.make(T.TypeError, state, None, f"Expected {expected_length} columns, but got {len(row)}")
         i = iter(row)
-        yield {name: restructure_result(state, col, i) for name, col in arr.type.elems.items()}
+        yield {name: restructure_result(state, kernel_type(col), i) for name, col in arr.type.elems.items()}
+
+@dp_type
+def restructure_result(state, t, i):
+    raise Signal.make(T.TypeError, None, f"Unexpected type used: {t}")
 
 @dp_type
 def restructure_result(state, t: T.table, i):
@@ -1062,19 +1182,15 @@ def restructure_result(state, t: T.table, i):
 
 @dp_type
 def restructure_result(state, t: T.struct, i):
-    return ({name: restructure_result(state, col, i) for name, col in t.elems.items()})
+    return ({name: restructure_result(state, kernel_type(col), i) for name, col in t.elems.items()})
 
 @dp_type
 def restructure_result(state, t: T.union[T.primitive, T.nulltype], i):
     return next(i)
 
-@dp_type
-def restructure_result(state, t: T.vectorized[T.union[T.primitive, T.nulltype]], i):
-    return next(i)
-
 
 @dp_type
-def restructure_result(state, t: T.list[T.union[T.primitive, T.nulltype]], i):
+def restructure_result(state, t: T.json_array[T.union[T.primitive, T.nulltype]], i):
     res = next(i)
     if not res:
         return res
@@ -1082,6 +1198,7 @@ def restructure_result(state, t: T.list[T.union[T.primitive, T.nulltype]], i):
     if state.db.target == mysql:
         res = json.loads(res)
     elif state.db.target == sqlite:
+        assert isinstance(res, str), res
         res = res.split(_ARRAY_SEP)
 
     # XXX hack! TODO Use a generic form to cast types
