@@ -79,6 +79,9 @@ class SqlInterface:
             table_type = self.import_table_type(state, table_name)
             yield None, table_name, table_type
 
+    def qualified_name(self, name):
+        return name
+
 
 
 def log_sql(sql):
@@ -216,20 +219,26 @@ class BigQueryInterface(SqlInterface):
         from google.cloud import bigquery
 
         self._client = bigquery.Client(project)
+        self._default_dataset = None
 
         self._print_sql = print_sql
 
 
     def table_exists(self, name):
-        # sql_code = "SELECT count(*) FROM information_schema.tables where table_name='%s'" % name
-        # cnt = self._execute_sql(T.int, sql_code, None)
-        # return cnt > 0
-        return False
+        from google.api_core.exceptions import NotFound
+        try:
+            self._client.get_table(name)
+        except NotFound:
+            return False
+        return True
+
+    def _list_tables(self):
+        for ds in self._client.list_datasets():
+            for t in self._client.list_tables(ds.reference):
+                yield t.full_table_id.replace(':', '.')     # Hacky
 
     def list_tables(self):
-        # sql_code = "SELECT table_name FROM information_schema.tables where table_schema='public'"
-        # return self._execute_sql(T.list[T.string], sql_code, None)
-        return []
+        return list(self._list_tables())
 
 
     def _execute_sql(self, sql_type, sql_code, state):
@@ -245,8 +254,6 @@ class BigQueryInterface(SqlInterface):
 
 
 
-    # Row(('bigquery-public-data', 'usa_names', 'usa_1910_2013', 'BASE TABLE', 'YES', 'NO', datetime.datetime(2016, 3, 12, 1, 2, 22, 425000, tzinfo=<UTC>)), {'table_catalog': 0, 'table_schema': 1, 'table_name': 2, 'table_type': 3, 'is_insertable_into': 4, 'is_typed': 5, 'creation_time': 6})
-    # Row(('bigquery-public-data', 'usa_names', 'usa_1910_current', 'BASE TABLE', 'YES', 'NO', datetime.datetime(2017, 2, 27, 21, 35, 45, 458000, tzinfo=<UTC>)), {'table_catalog': 0, 'table_schema': 1, 'table_name': 2, 'table_type': 3, 'is_insertable_into': 4, 'is_typed': 5, 'creation_time': 6})
     _schema_columns_t = T.table(dict(
         schema=T.string,
         table=T.string,
@@ -256,22 +263,39 @@ class BigQueryInterface(SqlInterface):
         type=T.string,
     ))
 
-    def import_table_types(self, state):
-        return []
-        # columns_q = """SELECT table_schema, table_name, column_name, ordinal_position, is_nullable, data_type
-        #     FROM information_schema.columns
-        #     """
-        # sql_columns = self._execute_sql(self._schema_columns_t, columns_q, state)
+    def import_table_type(self, state, name, columns_whitelist=None):
+        cols = {}
+        for f in self._client.get_table(name).schema:
+            if columns_whitelist is None or f.name in columns_whitelist:
+                cols[f.name] = _type_from_sql(f.field_type, f.is_nullable)
 
-        # columns_by_table = classify(sql_columns, lambda c: (c['schema'], c['table']))
+        return T.table(cols, name=Id(name))
 
-        # for (schema, table_name), columns in columns_by_table.items():
-        #     cols = [(c['pos'], c['name'], _type_from_sql(c['type'], c['nullable'])) for c in columns]
-        #     cols.sort()
-        #     cols = dict(c[1:] for c in cols)
+    def list_datasets(self):
+        return [ds.dataset_id for ds in self._client.list_datasets()]
 
-        #     # name = '%s.%s' % (schema, table_name)
-        #     yield schema, table_name, T.table(cols, name=Id(schema, table_name))
+    def set_default_dataset(self, dataset):
+        self._default_dataset = dataset
+
+    def get_default_dataset(self):
+        if self._default_dataset is None:
+            datasets = self.list_datasets()
+            if not datasets:
+                raise exceptions.Signal(T.ValueError, None, "No dataset found.")
+            self._default_dataset = datasets[0]
+        return self._default_dataset
+
+    def qualified_name(self, name):
+        if '.' in name: # already has dataset
+            return name
+        return self.get_default_dataset() + '.' + name
+
+    def rollback(self):
+        # XXX No error? No warning?
+        pass
+    def commit(self):
+        # XXX No error? No warning?
+        pass
 
 
 class SqliteInterface(SqlInterface):
